import ipaddress
import os
import socket
from datetime import datetime
from importlib import metadata

import dolphie.DataTypes as DataTypes
import dolphie.Modules.MetricManager as MetricManager
import requests
from dolphie.Modules.ManualException import ManualException
from dolphie.Modules.MySQL import Database
from dolphie.Modules.Queries import MySQLQueries
from dolphie.Widgets.new_version_modal import NewVersionModal
from packaging.version import parse as parse_version
from textual.app import App
from textual.widgets import Switch

try:
    __package_name__ = metadata.metadata(__package__ or __name__)["Name"]
    __version__ = metadata.version(__package__ or __name__)
except Exception:
    __package_name__ = "Dolphie"
    __version__ = "N/A"


class Dolphie:
    def __init__(self):
        self.app: App = None
        self.app_version = __version__
        self.tab_name = None
        self.tab_id = None

        # Config options
        self.user: str = None
        self.password: str = None
        self.host: str = None
        self.port: int = 3306
        self.socket: str = None
        self.ssl: dict = {}
        self.config_file: str = None
        self.host_cache_file: str = None
        self.host_setup_file: str = None
        self.debug: bool = False
        self.refresh_interval: int = 1
        self.show_idle_threads: bool = False
        self.show_trxs_only: bool = False
        self.show_additional_query_columns: bool = False
        self.sort_by_time_descending: bool = True
        self.heartbeat_table: str = None
        self.user_filter: str = None
        self.db_filter: str = None
        self.host_filter: str = None
        self.query_time_filter: str = 0
        self.query_filter: str = None
        self.host_setup_available_hosts: list = []
        self.host_cache: dict = {}
        self.host_cache_from_file: dict = {}
        self.startup_panels: str = None
        self.graph_marker: str = None

        self.reset_runtime_variables()

    def reset_runtime_variables(self):
        self.metric_manager = MetricManager.MetricManager()
        self.replica_manager = DataTypes.ReplicaManager()
        self.panels = DataTypes.Panels()

        # Set the graph switches to what they're currently selected to since we reset metric_manager
        if self.app:
            # Set the default panels based on startup_panels to be visible
            for panel in self.panels.all():
                setattr(getattr(self.panels, panel), "visible", False)
            for panel in self.startup_panels:
                setattr(getattr(self.panels, panel), "visible", True)

            switches = self.app.query(f".switch_container_{self.tab_id} Switch")
            for switch in switches:
                switch: Switch
                metric_instance_name = switch.name
                metric = switch.id

                metric_instance = getattr(self.metric_manager.metrics, metric_instance_name)
                metric_data: MetricManager.MetricData = getattr(metric_instance, metric)
                metric_data.visible = switch.value

        self.dolphie_start_time: datetime = datetime.now()
        self.worker_start_time: datetime = datetime.now()
        self.worker_previous_start_time: datetime = datetime.now()
        self.first_loop: bool = False
        self.polling_latency: float = 0
        self.refresh_latency: str = "0"
        self.read_only_status: str = None
        self.processlist_threads: dict = {}
        self.processlist_threads_snapshot: dict = {}
        self.lock_transactions: dict = {}
        self.ddl: list = []
        self.pause_refresh: bool = False
        self.previous_binlog_position: int = 0
        self.previous_replica_sbm: int = 0
        self.innodb_metrics: dict = {}
        self.disk_io_metrics: dict = {}
        self.global_variables: dict = {}
        self.lock_metrics: dict = {}
        self.global_status: dict = {}
        self.binlog_status: dict = {}
        self.replication_status: dict = {}
        self.replication_applier_status: dict = {}
        self.replica_lag_source: str = None
        self.replica_lag: int = None
        self.active_redo_logs: int = None
        self.mysql_host: str = None
        self.binlog_transaction_compression_percentage: int = None

        # Types of hosts
        self.galera_cluster: bool = False
        self.group_replication: bool = False
        self.innodb_cluster: bool = False
        self.innodb_cluster_read_replica: bool = False
        self.replicaset: bool = False
        self.aws_rds: bool = False
        self.mariadb: bool = False

        # These are for group replication in replication panel
        self.is_group_replication_primary: bool = False
        self.group_replication_members: dict = {}
        self.group_replication_data: dict = {}

        # Database connection global_variables
        # Main connection is used for Textual's worker thread so it can run asynchronous
        self.main_db_connection: Database = None
        # Secondary connection is for ad-hoc commands that are not a part of the worker thread
        self.secondary_db_connection: Database = None
        self.performance_schema_enabled: bool = False
        self.use_performance_schema: bool = True
        self.server_uuid: str = None
        self.mysql_version: str = None
        self.host_distro: str = None

    def check_for_update(self):
        # Query PyPI API to get the latest version
        try:
            url = f"https://pypi.org/pypi/{__package_name__}/json"
            response = requests.get(url, timeout=3)

            if response.status_code == 200:
                data = response.json()

                # Extract the latest version from the response
                latest_version = data["info"]["version"]

                # Compare the current version with the latest version
                if parse_version(latest_version) > parse_version(__version__):
                    self.app.push_screen(NewVersionModal(current_version=__version__, latest_version=latest_version))
        except Exception:
            pass

    def is_mysql_version_at_least(self, target, use_version=None):
        version = self.mysql_version
        if use_version:
            version = use_version

        parsed_source = parse_version(version)
        parsed_target = parse_version(target)

        return parsed_source >= parsed_target

    def db_connect(self):
        db_connection_args = {
            "app": self.app,
            "tab_name": self.tab_name,
            "host": self.host,
            "user": self.user,
            "password": self.password,
            "socket": self.socket,
            "port": self.port,
            "ssl": self.ssl,
        }
        self.main_db_connection = Database(**db_connection_args)
        self.secondary_db_connection = Database(**db_connection_args, save_connection_id=False)

        global_variables = self.main_db_connection.fetch_status_and_variables("variables")

        basedir = global_variables.get("basedir")
        aurora_version = global_variables.get("aurora_version")
        version = global_variables.get("version").lower()
        version_comment = global_variables.get("version_comment").lower()
        version_split = version.split(".")
        self.mysql_version = "%s.%s.%s" % (
            version_split[0],
            version_split[1],
            version_split[2].split("-")[0],
        )

        # Get proper host version and fork
        if "percona xtradb cluster" in version_comment:
            self.host_distro = "Percona XtraDB Cluster"
        elif "percona server" in version_comment:
            self.host_distro = "Percona Server"
        elif "mariadb cluster" in version_comment:
            self.host_distro = "MariaDB Cluster"
            self.mariadb = True
        elif "mariadb" in version_comment or "mariadb" in version:
            self.host_distro = "MariaDB"
            self.mariadb = True
        elif aurora_version:
            self.host_distro = "Amazon Aurora"
            self.aws_rds = True
        elif "rdsdb" in basedir:
            self.host_distro = "Amazon RDS"
            self.aws_rds = True
        else:
            self.host_distro = "MySQL"

        # For RDS, we will use the host specified to connect with since hostname isn't related to the endpoint
        if self.aws_rds:
            self.mysql_host = f"{self.host.split('.rds.amazonaws.com')[0]}:{self.port}"
        else:
            self.mysql_host = f"{global_variables.get('hostname')}:{self.port}"

        major_version = int(version_split[0])
        self.server_uuid = global_variables.get("server_uuid")
        if "MariaDB" in self.host_distro and major_version >= 10:
            self.server_uuid = global_variables.get("server_id")

        if global_variables.get("performance_schema") == "ON":
            self.performance_schema_enabled = True

        # Check to see if the host is in a Galera cluster
        galera_matches = any(key.startswith("wsrep_") for key in global_variables.keys())
        if galera_matches:
            self.galera_cluster = True

        # Check to get information on what cluster type it is
        if self.is_mysql_version_at_least("8.1"):
            query = MySQLQueries.determine_cluster_type_81
        else:
            query = MySQLQueries.determine_cluster_type_8

        self.main_db_connection.execute(query, ignore_error=True)
        data = self.main_db_connection.fetchone()

        cluster_type = data.get("cluster_type")
        instance_type = data.get("instance_type")

        if cluster_type == "ar":
            self.replicaset = True
        elif cluster_type == "gr":
            self.innodb_cluster = True

            if instance_type == "read-replica":
                self.innodb_cluster = False  # It doesn't work like an actual member in a cluster so set it to False
                self.innodb_cluster_read_replica = True

        if not self.innodb_cluster and global_variables.get("group_replication_group_name"):
            self.group_replication = True

        # Add host to host setup file if it doesn't exist
        with open(self.host_setup_file, "a+") as file:
            file.seek(0)
            lines = file.readlines()

            if self.port != 3306:
                host = f"{self.host}:{self.port}\n"
            else:
                host = f"{self.host}\n"

            if host not in lines:
                file.write(host)
                self.host_setup_available_hosts.append(host[:-1])  # remove the \n

    def monitor_read_only_change(self):
        current_ro_status = self.global_variables.get("read_only")
        formatted_ro_status = "RO" if current_ro_status == "ON" else "R/W"
        status = "read-only" if current_ro_status == "ON" else "read/write"

        message = f"Host [light_blue]{self.host}:{self.port}[/light_blue] is now [b highlight]{status}[/b highlight]"

        if current_ro_status == "ON" and not self.replication_status and not self.group_replication:
            message += " ([yellow]SHOULD BE READ/WRITE?[/yellow])"
        elif current_ro_status == "ON" and self.group_replication and self.is_group_replication_primary:
            message += " ([yellow]SHOULD BE READ/WRITE?[/yellow])"

        if self.read_only_status != formatted_ro_status and self.first_loop:
            self.app.notify(title="Read-only mode change", message=message, severity="warning", timeout=15)

        self.read_only_status = formatted_ro_status

    def command_input_to_variable(self, return_data):
        variable = return_data[0]
        value = return_data[1]
        if value:
            setattr(self, variable, value)

    def load_host_cache_file(self):
        if os.path.exists(self.host_cache_file):
            with open(self.host_cache_file) as file:
                for line in file:
                    line = line.strip()
                    error_message = f"Host cache entry '{line}' is not properly formatted! Format: ip=hostname"

                    if "=" not in line:
                        raise ManualException(error_message)

                    host, hostname = line.split("=", maxsplit=1)
                    host = host.strip()
                    hostname = hostname.strip()

                    if not host or not hostname:
                        raise ManualException(error_message)

                    self.host_cache_from_file[host] = hostname

    def get_hostname(self, host):
        if host in self.host_cache:
            return self.host_cache[host]

        if self.host_cache_from_file and host in self.host_cache_from_file:
            self.host_cache[host] = self.host_cache_from_file[host]
            return self.host_cache_from_file[host]

        try:
            ipaddress.IPv4Network(host)
            hostname = socket.gethostbyaddr(host)[0]
            self.host_cache[host] = hostname
        except (ValueError, socket.error):
            self.host_cache[host] = host
            hostname = host

        return hostname

    def massage_metrics_data(self):
        if self.is_mysql_version_at_least("8.0"):
            # If we're using MySQL 8, we need to fetch the checkpoint age from the performance schema if it's not
            # available in global status
            if not self.global_status.get("Innodb_checkpoint_age"):
                self.global_status["Innodb_checkpoint_age"] = self.main_db_connection.fetch_value_from_field(
                    MySQLQueries.checkpoint_age, "checkpoint_age"
                )

            if self.is_mysql_version_at_least("8.0.30"):
                active_redo_logs_count = self.main_db_connection.fetch_value_from_field(
                    MySQLQueries.active_redo_logs, "count"
                )
                self.global_status["Active_redo_log_count"] = active_redo_logs_count

        # If the server doesn't support Innodb_lsn_current, use Innodb_os_log_written instead
        # which has less precision, but it's good enough
        if not self.global_status.get("Innodb_lsn_current"):
            self.global_status["Innodb_lsn_current"] = self.global_status["Innodb_os_log_written"]

    def fetch_replication_data(self, replica: DataTypes.Replica = None):
        use_version = self.mysql_version
        if replica:
            use_version = replica.mysql_version

        if self.heartbeat_table:
            query = MySQLQueries.heartbeat_replica_lag
            replica_lag_source = "HB"
        elif (
            self.is_mysql_version_at_least("8.0", use_version) and self.performance_schema_enabled and not self.mariadb
        ):
            replica_lag_source = "PS"
            query = MySQLQueries.ps_replica_lag
        else:
            query = MySQLQueries.replication_status
            replica_lag_source = None

        replica_lag_data = None
        if replica:
            replica.connection.execute(query)
            replica_lag_data = replica.connection.fetchone()
        else:
            self.main_db_connection.execute(MySQLQueries.replication_status)
            replica_lag_data = self.main_db_connection.fetchone()
            self.replication_status = replica_lag_data

            if self.replication_status:
                # Use a better way to detect replication lag if available
                if replica_lag_source:
                    self.main_db_connection.execute(query)
                    replica_lag_data = self.main_db_connection.fetchone()

                # If we're using MySQL 8, fetch the replication applier status data
                self.replication_applier_status = None
                if (
                    self.is_mysql_version_at_least("8.0")
                    and self.panels.replication.visible
                    and self.global_variables.get("slave_parallel_workers", 0) > 1
                ):
                    self.main_db_connection.execute(MySQLQueries.replication_applier_status)
                    self.replication_applier_status = self.main_db_connection.fetchall()

        replica_lag = None
        if replica_lag_data and replica_lag_data["Seconds_Behind_Master"] is not None:
            replica_lag = int(replica_lag_data["Seconds_Behind_Master"])

        if replica:
            return replica_lag_source, replica_lag
        else:
            # Save the previous replica lag for to determine Speed data point
            self.previous_replica_sbm = self.replica_lag

            self.replica_lag_source = replica_lag_source
            self.replica_lag = replica_lag



import textwrap

from dolphie.Modules.MySQL import Database
from dolphie.Modules.Queries import MySQLQueries
from dolphie.Widgets.topbar import TopBar
from textual import events, on
from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.screen import Screen
from textual.widgets import DataTable, Input, Label, LoadingIndicator, Switch


class EventLog(Screen):
    CSS = """
        EventLog Horizontal {
            height: auto;
            align: center top;
            background: #0a0e1b;
            width: 100%;
        }
        EventLog Horizontal > Label {
            color: #bbc8e8;
            text-style: bold;
            margin-right: -1;
        }
        EventLog DataTable {
            background: #0a0e1b;
            border: none;
            overflow-x: hidden;
            max-height: 100%;
        }
        #info {
            padding-top: 1;
            width: 100%;
            text-align: center;
            text-style: bold;
        }
        #search {
            background: #0a0e1b;
            content-align: right middle;
            padding-left: 1;
            margin: 0;
            border: none;
        }
        #help {
            color: #8f9fc1;
            width: 100%;
            content-align: right middle;
        }
    """

    def __init__(self, read_only, app_version, host, db_connection: Database):
        super().__init__()

        self.read_only = read_only
        self.app_version = app_version
        self.host = host
        self.db_connection = db_connection

        self.levels = {
            "system": {"active": True, "sql": "prio = 'System'"},
            "warning": {"active": True, "sql": "prio = 'Warning'"},
            "error": {"active": True, "sql": "prio = 'Error'"},
        }

    def on_mount(self):
        datatable = self.query_one(DataTable)
        datatable.focus()
        datatable.styles.overflow_x = "auto"

        self.update_datatable()

    @on(events.Key)
    def on_keypress(self, event: events.Key):
        if event.key == "q":
            if self.screen.is_attached:
                self.app.pop_screen()
        elif event.key == "1":
            table = self.query_one(DataTable)
            table.move_cursor(row=0)
        elif event.key == "2":
            table = self.query_one(DataTable)
            table.move_cursor(row=table.row_count - 1)

    def compose(self) -> ComposeResult:
        yield TopBar(read_only=self.read_only, app_version=self.app_version, host=self.host)

        yield LoadingIndicator()

        yield Label("[b white]1[/b white] = top of events/[b white]2[/b white] = bottom of events", id="help")
        with Horizontal():
            switch_options = [("System", "system"), ("Warning", "warning"), ("Error", "error")]
            for label, switch_id in switch_options:
                yield Label(label)
                yield Switch(animate=False, id=switch_id, value=True)

        yield Label("", id="info")
        yield Input(id="search", placeholder="Search (hit enter when ready)")
        with Container():
            yield DataTable(show_cursor=False)

    @on(Input.Submitted, "#search")
    def event_search(self):
        self.update_datatable()

    @on(Switch.Changed)
    def switch_changed(self, event: Switch.Changed):
        self.levels[event.switch.id]["active"] = event.value

        self.update_datatable()

    def update_datatable(self):
        active_sql_list = [data["sql"] for data in self.levels.values() if data["active"]]
        where_clause = " OR ".join(active_sql_list)

        search_text = self.query_one("#search", Input)
        if search_text.value:
            where_clause = f"({where_clause}) AND (data LIKE '%{search_text.value}%')"

        table = self.query_one(DataTable)
        info = self.query_one("#info", Label)

        table.clear(columns=True)
        table.add_column("Time")
        table.add_column("Level")

        if where_clause:
            query = MySQLQueries.error_log.replace("$1", f"AND ({where_clause})")
            event_count = self.db_connection.execute(query)
            data = self.db_connection.fetchall()

            if data:
                table.add_column(f"Event ({event_count})")

                info.display = False
                table.display = True
                search_text.display = True

                for row in data:
                    level_color = ""
                    if row["level"] == "Error":
                        level_color = "red"
                    elif row["level"] == "Warning":
                        level_color = "yellow"

                    level = row["level"]
                    if level_color:
                        level = f"[{level_color}]{row['level']}[/{level_color}]"

                    timestamp = f"[#858A97]{row['timestamp'].strftime('%Y-%m-%d %H:%M:%S')}[/#858A97]"

                    # Wrap the message to 78% of console width so hopefully we don't get a scrollbar
                    wrapped_message = textwrap.wrap(row["message"], width=round(self.app.console.width * 0.78))
                    wrapped_message = "\n".join(wrapped_message)

                    line_counts = [cell.count("\n") + 1 for cell in wrapped_message]
                    height = max(line_counts)

                    table.add_row(timestamp, level, wrapped_message, height=height)
            else:
                table.display = False
                info.display = True
                info.update("No events found")
        else:
            table.display = False
            search_text.display = False
            info.display = True
            info.update("Toggle the switches above to filter what events you'd like to see")

        self.query_one(LoadingIndicator).display = False



from decimal import Decimal

import charset_normalizer


def format_bytes(bytes_value, color=True):
    units = ["B", "KB", "MB", "GB", "TB"]
    unit_index = 0

    while bytes_value >= 1024 and unit_index < len(units) - 1:
        bytes_value /= 1024
        unit_index += 1

    formatted_value = f"{bytes_value:.2f}"

    if formatted_value.endswith(".00"):
        formatted_value = formatted_value[:-3]  # Remove ".00" from the end

    if color:
        return f"{formatted_value}[highlight]{units[unit_index]}[/highlight]"
    else:
        return f"{formatted_value}{units[unit_index]}"


def format_time(time: int, picoseconds=False):
    if time is None:
        return "N/A"

    seconds = time / 1e12 if picoseconds else time

    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    return f"{int(hours):02d}:{int(minutes):02d}:{int(seconds):02d}"


def detect_encoding(text):
    # Since BLOB/BINARY data can be involved, we need to auto-detect what the encoding is
    # for queries since it can be anything. If I let pymsql use unicode by default I got
    # consistent crashes due to unicode errors for utf8 so we have to go this route
    result = charset_normalizer.detect(text)
    encoding = result["encoding"]

    if encoding is None:
        encoding = "latin1"
    elif encoding == "utf-16be":
        encoding = "utf-8"

    return encoding


def round_num(n, decimal=2):
    n = Decimal(n)
    return n.to_integral() if n == n.to_integral() else round(n.normalize(), decimal)


# This is from https://pypi.org/project/numerize
def format_number(n, decimal=2, color=True):
    if not n:
        return "0"

    # fmt: off
    sufixes = ["", "K", "M", "B", "T", "Qa", "Qu", "S", "Oc", "No",
               "D", "Ud", "Dd", "Td", "Qt", "Qi", "Se", "Od", "Nd", "V",
               "Uv", "Dv", "Tv", "Qv", "Qx", "Sx", "Ox", "Nx", "Tn", "Qa",
               "Qu", "S", "Oc", "No", "D", "Ud", "Dd", "Td", "Qt", "Qi",
               "Se", "Od", "Nd", "V", "Uv", "Dv", "Tv", "Qv", "Qx", "Sx",
               "Ox", "Nx", "Tn", "x", "xx", "xxx", "X", "XX", "XXX", "END"]

    sci_expr = [1e0, 1e3, 1e6, 1e9, 1e12, 1e15, 1e18, 1e21, 1e24, 1e27,
                1e30, 1e33, 1e36, 1e39, 1e42, 1e45, 1e48, 1e51, 1e54, 1e57,
                1e60, 1e63, 1e66, 1e69, 1e72, 1e75, 1e78, 1e81, 1e84, 1e87,
                1e90, 1e93, 1e96, 1e99, 1e102, 1e105, 1e108, 1e111, 1e114, 1e117,
                1e120, 1e123, 1e126, 1e129, 1e132, 1e135, 1e138, 1e141, 1e144, 1e147,
                1e150, 1e153, 1e156, 1e159, 1e162, 1e165, 1e168, 1e171, 1e174, 1e177]
    # fmt: on

    # Convert string to a number format if needed
    if isinstance(n, str):
        try:
            n = float(n)
        except ValueError:
            return n

    n = abs(n)
    for x in range(len(sci_expr)):
        if n >= sci_expr[x] and n < sci_expr[x + 1]:
            sufix = sufixes[x]
            if n >= 1e3:
                num = str(round_num(n / sci_expr[x], decimal))
            else:
                num = str(round_num(n, 0))
            if color:
                return f"{num}[highlight]{sufix}[/highlight]" if sufix else num
            else:
                return f"{num}{sufix}" if sufix else num


def format_sys_table_memory(data):
    parsed_data = data.strip().split(" ")
    if len(parsed_data) == 2:
        value, suffix = parsed_data[0], parsed_data[1][:1]

        if value == "0":
            suffix = ""
        elif suffix != "b":
            suffix += "B"
        elif suffix == "b":
            suffix = "B"

        return f"{value}[highlight]{suffix}"

    return data



from rich import box
from rich.syntax import Syntax
from rich.table import Table


class ManualException(Exception):
    def __init__(self, reason, query=""):
        self.reason = reason
        self.query = query

    def output(self):
        table_exception = Table(box=box.ROUNDED, show_header=True, style="#ec8888")

        table_exception.add_column("Error")
        if self.query:
            self.query = Syntax(
                self.query.strip(),
                "sql",
                line_numbers=False,
                word_wrap=True,
                theme="monokai",
                background_color="#121626",
            )
            table_exception.add_row("[white]Failed to execute query:[/white]")
            table_exception.add_row(self.query)
            table_exception.add_row("")

        if self.reason:
            # pymysql for some reason returns "ny connections" instead of "Too many connections"
            if isinstance(self.reason, str):
                self.reason = self.reason.replace("ny connections", "Too many connections")

            table_exception.add_row(self.reason)

        return table_exception



from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Label


class NewVersionModal(ModalScreen):
    CSS = """
        NewVersionModal > Vertical {
            background: #131626;
            border: tall #384673;
            height: auto;
            width: auto;
        }
        NewVersionModal > Vertical > * {
            width: auto;
            height: auto;
        }
        NewVersionModal #title {
            text-style: bold;
            width: 100%;
            margin: 0;
        }
        NewVersionModal Label {
            margin: 1 2;
            content-align: center middle;
        }
        NewVersionModal Horizontal {
            width: 100%;
            align-horizontal: center;
        }
    """
    BINDINGS = [
        Binding("escape", "app.pop_screen", "", show=False),
    ]

    def __init__(self, current_version, latest_version):
        super().__init__()

        self.current_version = current_version
        self.latest_version = latest_version

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label("[green]New version available!", id="title")
            yield Label(f"Current version: [highlight]{self.current_version}")
            yield Label(f"Latest version:  [highlight]{self.latest_version}")
            yield Label("Please update to the latest version at your convenience")
            yield Label("[dark_gray]Find more details at https://github.com/charles-001/dolphie")
            with Horizontal():
                yield Button("OK", variant="primary")

    def on_button_pressed(self) -> None:
        self.app.pop_screen()



#!/usr/bin/env python3

# ****************************
# *        Dolphie           *
# * Author: Charles Thompson *
# ****************************

import os
import re
import sys
from argparse import ArgumentParser, RawTextHelpFormatter
from configparser import ConfigParser
from datetime import datetime, timedelta
from functools import partial
from urllib.parse import urlparse

import dolphie.Modules.MetricManager as MetricManager
import myloginpath
from dolphie import Dolphie
from dolphie.DataTypes import ProcesslistThread
from dolphie.Modules.Functions import format_number, format_sys_table_memory
from dolphie.Modules.ManualException import ManualException
from dolphie.Modules.Queries import MySQLQueries
from dolphie.Modules.TabManager import Tab, TabManager
from dolphie.Panels import (
    dashboard_panel,
    ddl_panel,
    locks_panel,
    processlist_panel,
    replication_panel,
)
from dolphie.Widgets.command_screen import CommandScreen
from dolphie.Widgets.event_log_screen import EventLog
from dolphie.Widgets.modal import CommandModal
from dolphie.Widgets.thread_screen import ThreadScreen
from dolphie.Widgets.topbar import TopBar
from rich import box
from rich.align import Align
from rich.console import Console, Group
from rich.prompt import Prompt
from rich.style import Style
from rich.syntax import Syntax
from rich.table import Table
from rich.theme import Theme
from rich.traceback import Traceback
from sqlparse import format as sqlformat
from textual import events, on, work
from textual.app import App, ComposeResult
from textual.css.query import NoMatches
from textual.widgets import Switch, TabbedContent
from textual.worker import Worker, WorkerState, get_current_worker


def parse_args(dolphie: Dolphie):
    epilog = """
Config file with [client] section supports these options:
    host
    user
    password
    port
    socket
    ssl_mode REQUIRED/VERIFY_CA/VERIFY_IDENTITY
    ssl_ca
    ssl_cert
    ssl_key

Login path file supports these options:
    host
    user
    password
    port
    socket

Environment variables support these options:
    DOLPHIE_USER
    DOLPHIE_PASSWORD
    DOLPHIE_HOST
    DOLPHIE_PORT
    DOLPHIE_SOCKET

"""
    parser = ArgumentParser(
        conflict_handler="resolve",
        description="Dolphie, an intuitive feature-rich top tool for monitoring MySQL in real time",
        epilog=epilog,
        formatter_class=RawTextHelpFormatter,
    )

    parser.add_argument(
        "uri",
        metavar="uri",
        type=str,
        nargs="?",
        help=(
            "Use a URI string for credentials - format: mysql://user:password@host:port (port is optional with"
            " default 3306)"
        ),
    )

    parser.add_argument(
        "-u",
        "--user",
        dest="user",
        type=str,
        help="Username for MySQL",
    )
    parser.add_argument("-p", "--password", dest="password", type=str, help="Password for MySQL")
    parser.add_argument(
        "--ask-pass",
        dest="ask_password",
        action="store_true",
        default=False,
        help="Ask for password (hidden text)",
    )
    parser.add_argument(
        "-h",
        "--host",
        dest="host",
        type=str,
        help="Hostname/IP address for MySQL",
    )
    parser.add_argument(
        "-P",
        "--port",
        dest="port",
        type=int,
        help="Port for MySQL (Socket has precendence)",
    )
    parser.add_argument(
        "-S",
        "--socket",
        dest="socket",
        type=str,
        help="Socket file for MySQL",
    )
    parser.add_argument(
        "-c",
        "--config-file",
        dest="config_file",
        type=str,
        help=(
            "Config file path to use. This should use [client] section. "
            "See below for options support [default: ~/.my.cnf]"
        ),
    )
    parser.add_argument(
        "-f",
        "--host-cache-file",
        dest="host_cache_file",
        type=str,
        help=(
            "Resolve IPs to hostnames when your DNS is unable to. Each IP/hostname pair should be on its own line "
            "using format: ip=hostname [default: ~/dolphie_host_cache]"
        ),
    )
    parser.add_argument(
        "-q",
        "--host-setup-file",
        dest="host_setup_file",
        type=str,
        help=(
            "Specify location of file that stores the available hosts to use in host setup modal [default:"
            " ~/dolphie_hosts]"
        ),
    )
    parser.add_argument(
        "-l",
        "--login-path",
        dest="login_path",
        default="client",
        type=str,
        help=(
            "Specify login path to use mysql_config_editor's file ~/.mylogin.cnf for encrypted login credentials. "
            "Supercedes config file [default: %(default)s]"
        ),
    )
    parser.add_argument(
        "-r",
        "--refresh_interval",
        dest="refresh_interval",
        default=1,
        type=int,
        help="How much time to wait in seconds between each refresh [default: %(default)s]",
    )
    parser.add_argument(
        "-H",
        "--heartbeat-table",
        dest="heartbeat_table",
        type=str,
        help=(
            "If your hosts use pt-heartbeat, specify table in format db.table to use the timestamp it "
            "has for replication lag instead of Seconds_Behind_Master from SHOW SLAVE STATUS"
        ),
    )
    parser.add_argument(
        "--ssl-mode",
        dest="ssl_mode",
        type=str,
        help=(
            "Desired security state of the connection to the host. Supports: "
            "REQUIRED/VERIFY_CA/VERIFY_IDENTITY [default: OFF]"
        ),
    )
    parser.add_argument(
        "--ssl-ca",
        dest="ssl_ca",
        type=str,
        help="Path to the file that contains a PEM-formatted CA certificate",
    )
    parser.add_argument(
        "--ssl-cert",
        dest="ssl_cert",
        type=str,
        help="Path to the file that contains a PEM-formatted client certificate",
    )
    parser.add_argument(
        "--ssl-key",
        dest="ssl_key",
        type=str,
        help="Path to the file that contains a PEM-formatted private key for the client certificate",
    )
    parser.add_argument(
        "--panels",
        dest="startup_panels",
        default="dashboard,processlist",
        type=str,
        help=(
            "What panels to display on startup separated by a comma. Supports:"
            f" {'/'.join(dolphie.panels.all())} [default: %(default)s]"
        ),
    )
    parser.add_argument(
        "--graph-marker",
        dest="graph_marker",
        default="braille",
        type=str,
        help=(
            "What marker to use for graphs (available options: https://tinyurl.com/dolphie-markers) [default:"
            " %(default)s]"
        ),
    )
    parser.add_argument(
        "--show-trxs-only",
        dest="show_trxs_only",
        action="store_true",
        default=False,
        help="Start with only showing threads that have an active transaction",
    )
    parser.add_argument(
        "--additional-columns",
        dest="show_additional_query_columns",
        action="store_true",
        default=False,
        help="Start with additional columns in Processlist panel",
    )
    parser.add_argument(
        "--use-processlist",
        dest="use_processlist",
        action="store_true",
        default=False,
        help="Start with using Information Schema instead of Performance Schema for processlist panel",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=dolphie.app_version, help="Display version and exit"
    )

    console = Console(style="indian_red", highlight=False)

    home_dir = os.path.expanduser("~")

    parameter_options = vars(parser.parse_args())  # Convert object to dict
    basic_options = ["user", "password", "host", "port", "socket"]

    dolphie.config_file = f"{home_dir}/.my.cnf"
    if parameter_options["config_file"]:
        dolphie.config_file = parameter_options["config_file"]

    # Use config file for login credentials
    if os.path.isfile(dolphie.config_file):
        cfg = ConfigParser()
        cfg.read(dolphie.config_file)

        for option in basic_options:
            if cfg.has_option("client", option):
                setattr(dolphie, option, cfg.get("client", option))

        if cfg.has_option("client", "ssl_mode"):
            ssl_mode = cfg.get("client", "ssl_mode").upper()

            if ssl_mode == "REQUIRED":
                dolphie.ssl[""] = True
            elif ssl_mode == "VERIFY_CA":
                dolphie.ssl["check_hostname"] = False
            elif ssl_mode == "VERIFY_IDENTITY":
                dolphie.ssl["check_hostname"] = True
            else:
                sys.exit(console.print(f"Unsupported SSL mode [b]{ssl_mode}[/b]"))

        if cfg.has_option("client", "ssl_ca"):
            dolphie.ssl["ca"] = cfg.get("client", "ssl_ca")
        if cfg.has_option("client", "ssl_cert"):
            dolphie.ssl["cert"] = cfg.get("client", "ssl_cert")
        if cfg.has_option("client", "ssl_key"):
            dolphie.ssl["key"] = cfg.get("client", "ssl_key")

    # Use login path for login credentials
    if parameter_options["login_path"]:
        try:
            login_path_data = myloginpath.parse(parameter_options["login_path"])

            for option in basic_options:
                if option in login_path_data:
                    setattr(dolphie, option, login_path_data[option])
        except Exception as e:
            # Don't error out for default login path
            if parameter_options["login_path"] != "client":
                sys.exit(console.print(f"Problem reading login path file: {e}"))

    # Use environment variables for basic options if specified
    for option in basic_options:
        environment_var = "DOLPHIE_%s" % option.upper()
        if environment_var in os.environ and os.environ[environment_var]:
            setattr(dolphie, option, os.environ[environment_var])

    # Use parameter options if specified
    for option in basic_options:
        if parameter_options[option]:
            setattr(dolphie, option, parameter_options[option])

    # Lastly, parse URI if specified
    if parameter_options["uri"]:
        try:
            parsed = urlparse(parameter_options["uri"])

            if parsed.scheme != "mysql":
                sys.exit(
                    console.print("Invalid URI scheme: Only 'mysql' is supported (see --help for more information)")
                )

            dolphie.user = parsed.username
            dolphie.password = parsed.password
            dolphie.host = parsed.hostname
            dolphie.port = parsed.port or 3306
        except Exception as e:
            sys.exit(console.print(f"Invalid URI: {e} (see --help for more information)"))

    if parameter_options["ask_password"]:
        dolphie.password = Prompt.ask("[b #91abec]Password", password=True)

    if not dolphie.host:
        dolphie.host = "localhost"

    if parameter_options["refresh_interval"]:
        dolphie.refresh_interval = parameter_options["refresh_interval"]

    if parameter_options["heartbeat_table"]:
        pattern_match = re.search(r"^(\w+\.\w+)$", parameter_options["heartbeat_table"])
        if pattern_match:
            dolphie.heartbeat_table = parameter_options["heartbeat_table"]
            MySQLQueries.heartbeat_replica_lag = MySQLQueries.heartbeat_replica_lag.replace(
                "$1", dolphie.heartbeat_table
            )
        else:
            sys.exit(console.print("Your heartbeat table did not conform to the proper format: db.table"))

    if parameter_options["ssl_mode"]:
        ssl_mode = parameter_options["ssl_mode"].upper()

        if ssl_mode == "REQUIRED":
            dolphie.ssl[""] = True
        elif ssl_mode == "VERIFY_CA":
            dolphie.ssl["check_hostame"] = False
        elif ssl_mode == "VERIFY_IDENTITY":
            dolphie.ssl["check_hostame"] = True
        else:
            sys.exit(console.print(f"Unsupported SSL mode [b]{ssl_mode}[/b]"))

    if parameter_options["ssl_ca"]:
        dolphie.ssl["ca"] = parameter_options["ssl_ca"]
    if parameter_options["ssl_cert"]:
        dolphie.ssl["cert"] = parameter_options["ssl_cert"]
    if parameter_options["ssl_key"]:
        dolphie.ssl["key"] = parameter_options["ssl_key"]

    dolphie.host_cache_file = f"{home_dir}/dolphie_host_cache"
    if parameter_options["host_cache_file"]:
        dolphie.host_cache_file = parameter_options["host_cache_file"]

    dolphie.host_setup_file = f"{home_dir}/dolphie_hosts"
    if parameter_options["host_setup_file"]:
        dolphie.host_setup_file = parameter_options["host_setup_file"]

    dolphie.show_trxs_only = parameter_options["show_trxs_only"]
    dolphie.show_additional_query_columns = parameter_options["show_additional_query_columns"]

    if parameter_options["use_processlist"]:
        dolphie.use_performance_schema = False

    dolphie.startup_panels = parameter_options["startup_panels"].split(",")
    for panel in dolphie.startup_panels:
        if panel not in dolphie.panels.all():
            sys.exit(console.print(f"Panel '{panel}' is not valid (see --help for more information)"))

    dolphie.graph_marker = parameter_options["graph_marker"]

    if os.path.exists(dolphie.host_setup_file):
        with open(dolphie.host_setup_file, "r") as file:
            dolphie.host_setup_available_hosts = [line.strip() for line in file]


class DolphieApp(App):
    TITLE = "Dolphie"
    CSS_PATH = "Dolphie.css"

    def __init__(self, dolphie: Dolphie):
        super().__init__()

        dolphie.app = self
        self.dolphie = dolphie
        self.tab: Tab = None

        theme = Theme({
            "white": "#e9e9e9",
            "green": "#54efae",
            "yellow": "#f6ff8f",
            "red": "#fd8383",
            "purple": "#b565f3",
            "dark_gray": "#969aad",
            "highlight": "#91abec",
            "label": "#c5c7d2",
            "b label": "b #c5c7d2",
            "light_blue": "#bbc8e8",
            "b white": "b #e9e9e9",
            "b highlight": "b #91abec",
            "b red": "b #fd8383",
            "b light_blue": "b #bbc8e8",
            "panel_border": "#6171a6",
            "table_border": "#333f62",
        })
        self.console.push_theme(theme)
        self.console.set_window_title(self.TITLE)

    @work(thread=True, group="main")
    def run_worker_main(self, tab_id: int):
        tab = self.tab_manager.get_tab(tab_id)

        # Get our worker thread
        tab.worker = get_current_worker()
        tab.worker.name = tab_id

        # We have to queue the tab for removal because if we don't do it in the worker thread it will
        # be removed while the worker thread is running which will error due to objects not existing
        if tab.queue_for_removal:
            tab.disconnect(update_topbar=False)
            self.tab_manager.tabs.pop(tab.id, None)

            return

        dolphie = tab.dolphie
        try:
            if not dolphie.main_db_connection or not dolphie.main_db_connection.is_connected():
                tab.update_topbar(f"[[white]CONNECTING[/white]] {dolphie.host}:{dolphie.port}")
                tab.loading_indicator.display = True

                dolphie.db_connect()

            dolphie.worker_start_time = datetime.now()
            dolphie.polling_latency = (dolphie.worker_start_time - dolphie.worker_previous_start_time).total_seconds()
            dolphie.refresh_latency = round(dolphie.polling_latency - dolphie.refresh_interval, 2)
            dolphie.worker_previous_start_time = dolphie.worker_start_time

            dolphie.global_variables = dolphie.main_db_connection.fetch_status_and_variables("variables")
            dolphie.global_status = dolphie.main_db_connection.fetch_status_and_variables("status")
            dolphie.innodb_metrics = dolphie.main_db_connection.fetch_status_and_variables("innodb_metrics")

            if dolphie.performance_schema_enabled and dolphie.is_mysql_version_at_least("5.7"):
                find_replicas_query = MySQLQueries.ps_find_replicas
            else:
                find_replicas_query = MySQLQueries.pl_find_replicas

            dolphie.main_db_connection.execute(find_replicas_query)
            available_replicas = dolphie.main_db_connection.fetchall()
            # We update the replica ports used if the number of replicas have changed
            if len(available_replicas) != len(dolphie.replica_manager.available_replicas):
                dolphie.replica_manager.ports = {}

                dolphie.main_db_connection.execute(MySQLQueries.get_replicas)
                replica_data = dolphie.main_db_connection.fetchall()
                for row in replica_data:
                    dolphie.replica_manager.ports[row["Slave_UUID"]] = row["Port"]

                dolphie.replica_manager.available_replicas = available_replicas

            dolphie.main_db_connection.execute(MySQLQueries.ps_disk_io)
            dolphie.disk_io_metrics = dolphie.main_db_connection.fetchone()

            dolphie.fetch_replication_data()
            dolphie.massage_metrics_data()

            if dolphie.group_replication or dolphie.innodb_cluster:
                if dolphie.is_mysql_version_at_least("8.0.13"):
                    dolphie.main_db_connection.execute(MySQLQueries.group_replication_get_write_concurrency)
                    dolphie.group_replication_data = dolphie.main_db_connection.fetchone()

                dolphie.main_db_connection.execute(MySQLQueries.get_group_replication_members)
                dolphie.group_replication_members = dolphie.main_db_connection.fetchall()
                for member_role_data in dolphie.group_replication_members:
                    if (
                        member_role_data.get("MEMBER_ID") == dolphie.server_uuid
                        and member_role_data.get("MEMBER_ROLE") == "PRIMARY"
                    ):
                        dolphie.is_group_replication_primary = True
                        break

            if dolphie.panels.dashboard.visible:
                dolphie.main_db_connection.execute(MySQLQueries.binlog_status)
                dolphie.binlog_status = dolphie.main_db_connection.fetchone()

                # This can cause MySQL to crash: https://perconadev.atlassian.net/browse/PS-9066
                # if dolphie.global_variables.get("binlog_transaction_compression") == "ON":
                #     dolphie.main_db_connection.execute(MySQLQueries.get_binlog_transaction_compression_percentage)
                #     dolphie.binlog_transaction_compression_percentage = dolphie.main_db_connection.fetchone().get(
                #         "compression_percentage"
                #     )

            if dolphie.panels.processlist.visible:
                dolphie.processlist_threads = processlist_panel.fetch_data(tab)

            if dolphie.is_mysql_version_at_least("5.7"):
                # We don't check if panel is visible or not since we use this data for Locks graph
                dolphie.main_db_connection.execute(MySQLQueries.locks_query)
                dolphie.lock_transactions = dolphie.main_db_connection.fetchall()

                if dolphie.panels.ddl.visible:
                    dolphie.main_db_connection.execute(MySQLQueries.ddls)
                    dolphie.ddl = dolphie.main_db_connection.fetchall()

            dolphie.monitor_read_only_change()

            dolphie.metric_manager.refresh_data(
                worker_start_time=dolphie.worker_start_time,
                polling_latency=dolphie.polling_latency,
                global_variables=dolphie.global_variables,
                global_status=dolphie.global_status,
                innodb_metrics=dolphie.innodb_metrics,
                disk_io_metrics=dolphie.disk_io_metrics,
                lock_metrics=dolphie.lock_transactions,
                replication_status=dolphie.replication_status,
                replication_lag=dolphie.replica_lag,
            )
        except ManualException as e:
            # This will set up the worker state change function below to trigger the
            # host setup modal with the error
            tab.worker_cancel_error = e.output()

            tab.disconnect()

    def on_worker_state_changed(self, event: Worker.StateChanged):
        tab = self.tab_manager.get_tab(event.worker.name)
        if not tab:
            return

        dolphie = tab.dolphie

        if event.worker.group == "main":
            if event.state == WorkerState.SUCCESS:
                # Skip this if the conditions are right
                if (
                    len(self.screen_stack) > 1
                    or dolphie.pause_refresh
                    or not tab.dolphie.main_db_connection.is_connected()
                    or tab.id != self.tab.id
                ):
                    tab.worker_timer = self.set_timer(
                        tab.dolphie.refresh_interval, partial(self.run_worker_main, tab.id)
                    )

                    return

                self.refresh_screen(tab)

                tab.worker_timer = self.set_timer(tab.dolphie.refresh_interval, partial(self.run_worker_main, tab.id))
            elif event.state == WorkerState.CANCELLED:
                # Only show the modal if there's a worker cancel error
                if tab.worker_cancel_error:
                    if self.tab.id != tab.id:
                        self.notify(
                            f"Host [light_blue]{tab.dolphie.host}:{tab.dolphie.port}[/light_blue] has been"
                            " disconnected",
                            title="Error",
                            severity="error",
                            timeout=10,
                        )

                    self.tab_manager.switch_tab(tab.id)

                    tab.host_setup()
                    self.bell()
        elif event.worker.group == "replicas":
            if event.state == WorkerState.SUCCESS:
                # Skip this if the conditions are right
                if len(self.screen_stack) > 1 or dolphie.pause_refresh or tab.id != self.tab.id:
                    tab.replicas_worker_timer = self.set_timer(
                        tab.dolphie.refresh_interval, partial(self.run_worker_replicas, tab.id)
                    )

                    return

                if dolphie.panels.replication.visible and dolphie.replica_manager.available_replicas:
                    replication_panel.create_replica_panel(tab)

                tab.replicas_worker_timer = self.set_timer(2, partial(self.run_worker_replicas, tab.id))

    @work(thread=True, group="replicas")
    def run_worker_replicas(self, tab_id: int):
        tab = self.tab_manager.get_tab(tab_id)

        # Get our worker thread
        tab.replicas_worker = get_current_worker()
        tab.replicas_worker.name = tab_id

        dolphie = tab.dolphie

        if dolphie.panels.replication.visible:
            if dolphie.replica_manager.available_replicas:
                if not dolphie.replica_manager.replicas:
                    tab.replicas_container.display = True
                    tab.replicas_loading_indicator.display = True
                    tab.replicas_title.update(
                        f"[b]Loading [highlight]{len(dolphie.replica_manager.available_replicas)}[/highlight]"
                        " replicas...\n"
                    )

                replication_panel.fetch_replicas(tab)
            else:
                tab.replicas_container.display = False
        else:
            # If we're not displaying the replication panel, close all replica connections
            dolphie.replica_manager.remove_all()

    def refresh_screen(self, tab: Tab):
        dolphie = tab.dolphie

        try:
            loading_indicator = tab.loading_indicator
            if loading_indicator.display:
                # We only want to do this on startup
                # Set panels to be visible for the ones the user specifies
                for panel in dolphie.panels.all():
                    setattr(getattr(tab, f"panel_{panel}"), "display", False)

                for panel in dolphie.startup_panels:
                    setattr(getattr(tab, f"panel_{panel}"), "display", True)

                loading_indicator.display = False
                tab.main_container.display = True

                self.layout_graphs()

            if self.tab.id == tab.id:
                tab.update_topbar()

            if dolphie.panels.dashboard.visible:
                self.refresh_panel(dolphie.panels.dashboard.name)

                # Update the sparkline for queries per second
                sparkline = tab.sparkline
                sparkline_data = dolphie.metric_manager.metrics.dml.Queries.values
                if not sparkline.display:
                    sparkline_data = [0]
                    sparkline.display = True

                sparkline.data = sparkline_data
                sparkline.refresh()

            if dolphie.panels.processlist.visible:
                self.refresh_panel(dolphie.panels.processlist.name)

            if dolphie.panels.replication.visible:
                self.refresh_panel(dolphie.panels.replication.name)

            if dolphie.panels.locks.visible:
                self.refresh_panel(dolphie.panels.locks.name)

            if dolphie.panels.ddl.visible:
                self.refresh_panel(dolphie.panels.ddl.name)

            if dolphie.panels.graphs.visible:
                graph_panel = self.query_one(f"#tabbed_content_{tab.id}", TabbedContent)

                # Hide/show replication tab based on replication status
                if dolphie.replication_status:
                    graph_panel.show_tab(f"graph_tab_replication_lag_{tab.id}")
                else:
                    graph_panel.hide_tab(f"graph_tab_replication_lag_{tab.id}")

                # Refresh the graph(s) for the selected tab
                self.update_graphs(graph_panel.get_pane(graph_panel.active).name)

            # We take a snapshot of the processlist to be used for commands
            # since the data can change after a key is pressed
            dolphie.processlist_threads_snapshot = dolphie.processlist_threads.copy()

            # This denotes that we've gone through the first loop of the worker thread
            dolphie.first_loop = True
        except NoMatches:
            # This is thrown if a user toggles panels on and off and the display_* states aren't 1:1
            # with worker thread/state change due to asynchronous nature of the worker thread
            pass

    async def on_key(self, event: events.Key):
        if len(self.screen_stack) > 1:
            return

        await self.capture_key(event.key)

    async def on_mount(self):
        dolphie = self.dolphie

        self.tab_manager = TabManager(self)
        await self.tab_manager.create_tab("Main", dolphie)

        dolphie.load_host_cache_file()
        dolphie.check_for_update()

        self.run_worker_main(self.tab.id)
        self.run_worker_replicas(self.tab.id)

    def _handle_exception(self, error: Exception) -> None:
        self.bell()
        self.exit(message=Traceback(show_locals=True, width=None, locals_max_length=5))

    @on(TabbedContent.TabActivated, "#tabbed_content")
    def tab_changed(self, event: TabbedContent.TabActivated):
        self.tab_manager.switch_tab(event.pane.name)

        if self.tab.dolphie.main_db_connection:
            self.refresh_screen(self.tab)

    @on(TabbedContent.TabActivated, ".metrics_tabbed_content")
    def metric_tab_changed(self, event: TabbedContent.TabActivated):
        metric_instance_name = event.pane.name

        if metric_instance_name:
            self.update_graphs(metric_instance_name)

    def update_graphs(self, tab_metric_instance_name):
        if not self.tab.panel_graphs.display:
            return

        for metric_instance in self.tab.dolphie.metric_manager.metrics.__dict__.values():
            if tab_metric_instance_name == metric_instance.tab_name:
                for graph_name in metric_instance.graphs:
                    self.query_one(f"#{graph_name}_{self.tab.id}").render_graph(metric_instance)

        self.update_stats_label(tab_metric_instance_name)

    def update_stats_label(self, tab_metric_instance_name):
        stat_data = {}

        for metric_instance in self.tab.dolphie.metric_manager.metrics.__dict__.values():
            if hasattr(metric_instance, "tab_name") and metric_instance.tab_name == tab_metric_instance_name:
                number_format_func = MetricManager.get_number_format_function(metric_instance, color=True)
                for metric_data in metric_instance.__dict__.values():
                    if isinstance(metric_data, MetricManager.MetricData) and metric_data.values and metric_data.visible:
                        stat_data[metric_data.label] = number_format_func(metric_data.values[-1])

        formatted_stat_data = "  ".join(
            f"[b light_blue]{label}[/b light_blue] {value}" for label, value in stat_data.items()
        )
        self.query_one(f"#stats_{tab_metric_instance_name}_{self.tab.id}").update(formatted_stat_data)

    def toggle_panel(self, panel_name):
        panel = self.app.query_one(f"#panel_{panel_name}_{self.tab.id}")

        new_display = not panel.display
        panel.display = new_display

        setattr(getattr(self.tab.dolphie.panels, panel_name), "visible", new_display)

        if panel_name not in [self.tab.dolphie.panels.graphs.name]:
            self.app.refresh_panel(panel_name, toggled=True)

    def refresh_panel(self, panel_name, toggled=False):
        # If loading indicator is displaying, don't refresh
        if self.tab.loading_indicator.display:
            return

        if panel_name == self.tab.dolphie.panels.replication.name:
            # When replication panel status is changed, we need to refresh the dashboard panel as well since
            # it adds/removes it from there
            replication_panel.create_panel(self.tab)

            if toggled and self.tab.dolphie.replication_status:
                dashboard_panel.create_panel(self.tab)
        elif panel_name == self.tab.dolphie.panels.dashboard.name:
            dashboard_panel.create_panel(self.tab)
        elif panel_name == self.tab.dolphie.panels.processlist.name:
            processlist_panel.create_panel(self.tab)
        elif panel_name == self.tab.dolphie.panels.locks.name:
            locks_panel.create_panel(self.tab)
        elif panel_name == self.tab.dolphie.panels.ddl.name:
            ddl_panel.create_panel(self.tab)

        if toggled or not self.tab.dolphie.first_loop:  # Include this so we're not updating the sizes every loop
            # Update the sizes of the panels depending if replication container is visible or not
            if self.tab.dolphie.replication_status and not self.tab.dolphie.panels.replication.visible:
                self.tab.dashboard_host_information.styles.width = "25vw"
                self.tab.dashboard_binary_log.styles.width = "21vw"

                self.tab.dashboard_innodb.styles.width = "17vw"
                self.tab.dashboard_replication.styles.width = "25vw"
                self.tab.dashboard_statistics.styles.width = "12vw"
            else:
                self.tab.dashboard_host_information.styles.width = "32vw"
                self.tab.dashboard_binary_log.styles.width = "27vw"
                self.tab.dashboard_innodb.styles.width = "24vw"
                self.tab.dashboard_replication.styles.width = "0"
                self.tab.dashboard_statistics.styles.width = "17vw"

            self.tab.dashboard_host_information.styles.max_width = "45"
            self.tab.dashboard_binary_log.styles.max_width = "38"
            self.tab.dashboard_innodb.styles.max_width = "32"
            self.tab.dashboard_replication.styles.max_width = "50"
            self.tab.dashboard_statistics.styles.max_width = "22"

    def layout_graphs(self):
        if self.tab.dolphie.is_mysql_version_at_least("8.0.30"):
            self.query_one(f"#graph_redo_log_{self.tab.id}").styles.width = "55%"
            self.query_one(f"#graph_redo_log_bar_{self.tab.id}").styles.width = "12%"
            self.query_one(f"#graph_redo_log_active_count_{self.tab.id}").styles.width = "33%"
            self.tab.dolphie.metric_manager.metrics.redo_log_active_count.Active_redo_log_count.visible = True
            self.query_one(f"#graph_redo_log_active_count_{self.tab.id}").display = True
        else:
            self.query_one(f"#graph_redo_log_{self.tab.id}").styles.width = "88%"
            self.query_one(f"#graph_redo_log_bar_{self.tab.id}").styles.width = "12%"
            self.query_one(f"#graph_redo_log_active_count_{self.tab.id}").display = False

        self.query_one(f"#graph_adaptive_hash_index_{self.tab.id}").styles.width = "50%"
        self.query_one(f"#graph_adaptive_hash_index_hit_ratio_{self.tab.id}").styles.width = "50%"

    @on(Switch.Changed)
    def switch_changed(self, event: Switch.Changed):
        if len(self.screen_stack) > 1:
            return

        metric_instance_name = event.switch.name
        metric = event.switch.id

        metric_instance = getattr(self.tab.dolphie.metric_manager.metrics, metric_instance_name)
        metric_data: MetricManager.MetricData = getattr(metric_instance, metric)
        metric_data.visible = event.value

        self.update_graphs(metric_instance_name)

    async def capture_key(self, key):
        tab = self.tab_manager.get_tab(self.tab.id)

        exclude_keys = [
            "up",
            "down",
            "left",
            "right",
            "pageup",
            "pagedown",
            "home",
            "end",
            "tab",
            "enter",
            "grave_accent",
            "q",
            "question_mark",
            "plus",
            "minus",
            "ctrl+a",
            "ctrl+d",
        ]

        screen_data = None
        dolphie = tab.dolphie

        # Prevent commands from being run if the secondary connection is processing a query already
        if key not in exclude_keys:
            if dolphie.secondary_db_connection and dolphie.secondary_db_connection.running_query:
                self.notify("There's already a command running - please wait for it to finish")
                return

            if not tab.worker:
                self.notify("You must be connected to a host to use commands")
                return

        if key == "1":
            self.toggle_panel(dolphie.panels.dashboard.name)
        elif key == "2":
            self.toggle_panel(dolphie.panels.processlist.name)
            self.tab.processlist_datatable.clear()
        elif key == "3":
            self.toggle_panel(dolphie.panels.replication.name)

            tab.replicas_container.display = False
            if not dolphie.panels.replication.visible:
                for member in dolphie.app.query(f".replica_container_{dolphie.tab_id}"):
                    member.remove()
            else:
                if dolphie.replica_manager.available_replicas:
                    tab.replicas_container.display = True
                    tab.replicas_title.update(
                        f"[b]Loading [highlight]{len(dolphie.replica_manager.available_replicas)}[/highlight]"
                        " replicas...\n"
                    )
                    tab.replicas_loading_indicator.display = True

        elif key == "4":
            self.toggle_panel(dolphie.panels.graphs.name)
            self.app.update_graphs("dml")
        elif key == "5":
            if not dolphie.is_mysql_version_at_least("5.7"):
                self.notify("Locks panel requires MySQL 5.7+")
                return

            self.toggle_panel(dolphie.panels.locks.name)
            self.tab.locks_datatable.clear()
        elif key == "6":
            if not dolphie.is_mysql_version_at_least("5.7"):
                self.notify("DDL panel requires MySQL 5.7+")
                return

            query = "SELECT enabled FROM performance_schema.setup_instruments WHERE name LIKE 'stage/innodb/alter%';"
            dolphie.secondary_db_connection.execute(query)
            data = dolphie.secondary_db_connection.fetchall()
            for row in data:
                if row.get("enabled") == "NO":
                    self.notify("DDL panel requires Performance Schema to have 'stage/innodb/alter%' enabled")
                    return

            self.toggle_panel(dolphie.panels.ddl.name)
            self.tab.ddl_datatable.clear()
        elif key == "grave_accent":
            self.tab.host_setup()
        elif key == "space":
            if tab.worker.state != WorkerState.RUNNING:
                tab.worker_timer.stop()
                self.run_worker_main(tab.id)
        elif key == "plus":

            async def command_get_input(tab_name):
                await self.tab_manager.create_tab(tab_name, dolphie)
                self.tab.topbar.host = ""
                self.tab.host_setup()

            self.app.push_screen(
                CommandModal(command="new_tab", message="What would you like to name the new tab?"),
                command_get_input,
            )
        elif key == "minus":
            if tab.id == 1:
                self.notify("Main tab cannot be removed")
                return
            else:
                await self.tab_manager.remove_tab(tab.id)
        elif key == "ctrl+a" or key == "ctrl+d":
            all_tabs = self.tab_manager.get_all_tabs()

            if key == "ctrl+a":
                switch_to_tab = all_tabs[(all_tabs.index(tab.id) - 1) % len(all_tabs)]
            elif key == "ctrl+d":
                switch_to_tab = all_tabs[(all_tabs.index(tab.id) + 1) % len(all_tabs)]

            self.tab_manager.switch_tab(switch_to_tab)

        elif key == "a":
            if dolphie.show_additional_query_columns:
                dolphie.show_additional_query_columns = False
                self.notify("Processlist will now hide additional columns")
            else:
                dolphie.show_additional_query_columns = True
                self.notify("Processlist will now show additional columns")

        elif key == "c":
            dolphie.user_filter = ""
            dolphie.db_filter = ""
            dolphie.host_filter = ""
            dolphie.query_time_filter = ""
            dolphie.query_filter = ""

            self.notify("Cleared all filters", severity="success")

        elif key == "d":
            self.run_command_in_worker(key=key, dolphie=dolphie)

        elif key == "D":
            tab.disconnect()

        elif key == "e":
            if dolphie.is_mysql_version_at_least("8.0") and dolphie.performance_schema_enabled:
                self.app.push_screen(
                    EventLog(
                        dolphie.read_only_status,
                        dolphie.app_version,
                        dolphie.mysql_host,
                        dolphie.secondary_db_connection,
                    )
                )
            else:
                self.notify("Error log command requires MySQL 8+ with Performance Schema enabled")

        elif key == "f":

            def command_get_input(filter_data):
                filter_name, filter_value = filter_data[0], filter_data[1]
                filters_mapping = {
                    "User": "user_filter",
                    "Database": "db_filter",
                    "Host": "host_filter",
                    "Query time": "query_time_filter",
                    "Query text": "query_filter",
                }

                attribute = filters_mapping.get(filter_name)
                if attribute:
                    setattr(dolphie, attribute, int(filter_value) if attribute == "query_time_filter" else filter_value)
                    self.notify(
                        f"Filtering [b]{filter_name.capitalize()}[/b] by [b highlight]{filter_value}[/b highlight]",
                        severity="success",
                    )
                else:
                    self.notify(f"Invalid filter name {filter_name}", severity="error")

            self.app.push_screen(
                CommandModal(
                    command="thread_filter",
                    message="Select which field you'd like to filter by",
                    processlist_data=dolphie.processlist_threads_snapshot,
                    host_cache_data=dolphie.host_cache,
                ),
                command_get_input,
            )

        elif key == "i":
            if dolphie.show_idle_threads:
                dolphie.show_idle_threads = False
                dolphie.sort_by_time_descending = True

                self.notify("Processlist will now hide idle threads")
            else:
                dolphie.show_idle_threads = True
                dolphie.sort_by_time_descending = False

                self.notify("Processlist will now show idle threads")

        elif key == "k":

            def command_get_input(data):
                self.run_command_in_worker(key=key, dolphie=dolphie, additional_data=data)

            self.app.push_screen(
                CommandModal(
                    command="thread_kill_by_id",
                    message="Specify a Thread ID to kill",
                    processlist_data=dolphie.processlist_threads_snapshot,
                ),
                command_get_input,
            )

        elif key == "K":

            def command_get_input(data):
                self.run_command_in_worker(key=key, dolphie=dolphie, additional_data=data)

            self.app.push_screen(
                CommandModal(
                    command="thread_kill_by_parameter",
                    message="Kill threads based around parameters",
                    processlist_data=dolphie.processlist_threads_snapshot,
                ),
                command_get_input,
            )

        elif key == "l":
            self.run_command_in_worker(key=key, dolphie=dolphie)

        elif key == "o":
            self.run_command_in_worker(key=key, dolphie=dolphie)

        elif key == "m":
            if not dolphie.is_mysql_version_at_least("5.7") or not dolphie.performance_schema_enabled:
                self.notify("Memory usage command requires MySQL 5.7+ with Performance Schema enabled")
            else:
                self.run_command_in_worker(key=key, dolphie=dolphie)

        elif key == "p":
            if not dolphie.pause_refresh:
                dolphie.pause_refresh = True
                self.notify(f"Refresh is paused! Press [b highlight]{key}[/b highlight] again to resume")
            else:
                dolphie.pause_refresh = False
                self.notify("Refreshing has resumed", severity="success")

        if key == "P":
            if dolphie.use_performance_schema:
                dolphie.use_performance_schema = False
                self.notify("Switched to using [b highlight]Processlist")
            else:
                if dolphie.performance_schema_enabled:
                    dolphie.use_performance_schema = True
                    self.notify("Switched to using [b highlight]Performance Schema")
                else:
                    self.notify("You can't switch to Performance Schema because it isn't enabled")

        elif key == "q":
            self.app.exit()

        elif key == "r":

            def command_get_input(refresh_interval):
                dolphie.refresh_interval = refresh_interval

                self.notify(
                    f"Refresh interval set to [b highlight]{refresh_interval}[/b highlight] second(s)",
                    severity="success",
                )

            self.app.push_screen(
                CommandModal(command="refresh_interval", message="Specify refresh interval (in seconds)"),
                command_get_input,
            )

        elif key == "R":
            dolphie.metric_manager.reset()

            active_graph = self.app.query_one(f"#tabbed_content_{tab.id}", TabbedContent)
            self.update_graphs(active_graph.get_pane(active_graph.active).name)
            self.notify("Metrics have been reset", severity="success")

        elif key == "s":
            if dolphie.sort_by_time_descending:
                dolphie.sort_by_time_descending = False
                self.notify("Processlist will now sort threads by time in ascending order")
            else:
                dolphie.sort_by_time_descending = True
                self.notify("Processlist will now sort threads by time in descending order")

        elif key == "t":

            def command_get_input(data):
                self.run_command_in_worker(key=key, dolphie=dolphie, additional_data=data)

            self.app.push_screen(
                CommandModal(
                    command="show_thread",
                    message="Specify a Thread ID to display its details",
                    processlist_data=dolphie.processlist_threads_snapshot,
                ),
                command_get_input,
            )

        elif key == "T":
            if dolphie.show_trxs_only:
                dolphie.show_trxs_only = False
                dolphie.show_idle_threads = False
                self.notify("Processlist will now no longer only show threads that have an active transaction")
            else:
                dolphie.show_trxs_only = True
                dolphie.show_idle_threads = True
                self.notify("Processlist will now only show threads that have an active transaction")

        elif key == "u":
            if not dolphie.performance_schema_enabled:
                self.notify("User statistics command requires Performance Schema to be enabled")
            else:
                self.run_command_in_worker(key=key, dolphie=dolphie)

        elif key == "v":

            def command_get_input(input_variable):
                table_grid = Table.grid()
                table_counter = 1
                variable_counter = 1
                row_counter = 1
                variable_num = 1
                all_tables = []
                tables = {}
                display_global_variables = {}

                for variable, value in dolphie.global_variables.items():
                    if input_variable == "all":
                        display_global_variables[variable] = dolphie.global_variables[variable]
                    else:
                        if input_variable:
                            if input_variable in variable:
                                display_global_variables[variable] = dolphie.global_variables[variable]

                max_num_tables = 1 if len(display_global_variables) <= 50 else 2

                # Create the number of tables we want
                while table_counter <= max_num_tables:
                    tables[table_counter] = Table(box=box.HORIZONTALS, show_header=False, style="table_border")
                    tables[table_counter].add_column("")
                    tables[table_counter].add_column("")

                    table_counter += 1

                # Calculate how many global_variables per table
                row_per_count = len(display_global_variables) // max_num_tables

                # Loop global_variables
                for variable, value in display_global_variables.items():
                    tables[variable_num].add_row("[label]%s" % variable, str(value))

                    if variable_counter == row_per_count and row_counter != max_num_tables:
                        row_counter += 1
                        variable_counter = 0
                        variable_num += 1

                    variable_counter += 1

                # Put all the variable data from dict into an array
                all_tables = [table_data for table_data in tables.values() if table_data]

                # Add the data into a single tuple for add_row
                if display_global_variables:
                    table_grid.add_row(*all_tables)
                    screen_data = Align.center(table_grid)

                    self.app.push_screen(
                        CommandScreen(dolphie.read_only_status, dolphie.app_version, dolphie.mysql_host, screen_data)
                    )
                else:
                    if input_variable:
                        self.notify("No variable(s) found that match [b highlight]%s[/b highlight]" % input_variable)

            self.app.push_screen(
                CommandModal(
                    command="variable_search",
                    message="Specify a variable to wildcard search\n[dim](input all to show everything)[/dim]",
                ),
                command_get_input,
            )

        elif key == "z":
            if dolphie.host_cache:
                table = Table(
                    box=box.SIMPLE_HEAVY,
                    show_edge=False,
                    style="table_border",
                )
                table.add_column("Host/IP")
                table.add_column("Hostname (if resolved)")

                for ip, addr in dolphie.host_cache.items():
                    if ip:
                        table.add_row(ip, addr)

                screen_data = Group(
                    Align.center(
                        "[b light_blue]Host Cache[/b light_blue] ([b highlight]%s[/b highlight])\n"
                        % len(dolphie.host_cache)
                    ),
                    table,
                )
            else:
                screen_data = Group(
                    Align.center("[b light_blue]Host Cache[/b light_blue]\n"), "There are currently no hosts resolved"
                )

        elif key == "question_mark":
            keys = {
                "1": "Show/hide Dashboard",
                "2": "Show/hide Processlist",
                "3": "Show/hide Replication/Replicas",
                "4": "Show/hide Graph Metrics",
                "5": "Show/hide Locks",
                "6": "Show/hide DDLs",
                "`": "Open Host Setup",
                "+": "Create a new tab",
                "-": "Remove the current tab",
                "a": "Toggle additional processlist columns",
                "c": "Clear all filters set",
                "d": "Display all databases",
                "D": "Disconnect from the tab's host",
                "e": "Display error log from Performance Schema",
                "f": "Filter processlist by a supported option",
                "i": "Toggle displaying idle threads",
                "k": "Kill a thread by its ID",
                "K": "Kill a thread by a supported option",
                "l": "Display the most recent deadlock",
                "o": "Display output from SHOW ENGINE INNODB STATUS",
                "m": "Display memory usage",
                "p": "Pause refreshing",
                "P": "Switch between using Information Schema/Performance Schema for processlist panel",
                "q": "Quit",
                "r": "Set the refresh interval",
                "R": "Reset all metrics",
                "t": "Display details of a thread along with an EXPLAIN of its query",
                "T": "Transaction view - toggle displaying threads that only have an active transaction",
                "s": "Sort processlist by time in descending/ascending order",
                "u": "List active connected users and their statistics",
                "v": "Variable wildcard search sourced from SHOW GLOBAL VARIABLES",
                "z": "Display all entries in the host cache",
                "space": "Force a manual refresh of all panels",
                "ctrl+a": "Switch to the previous tab",
                "ctrl+d": "Switch to the next tab",
            }

            table_keys = Table(
                box=box.SIMPLE_HEAVY,
                show_edge=False,
                style="table_border",
                title="Commands",
                title_style="bold #bbc8e8",
                header_style="bold",
            )
            table_keys.add_column("Key", justify="center", style="b highlight")
            table_keys.add_column("Description")

            for key, description in keys.items():
                table_keys.add_row(key, description)

            datapoints = {
                "Read Only": "If the host is in read-only mode",
                "Read Hit": "The percentage of how many reads are from InnoDB buffer pool compared to from disk",
                "Lag": (
                    "Retrieves metric from: Default -> SHOW SLAVE STATUS, HB -> Heartbeat table, PS -> Performance"
                    " Schema"
                ),
                "Chkpt Age": (
                    "This depicts how close InnoDB is before it starts to furiously flush dirty data to disk "
                    "(Lower is better)"
                ),
                "AHI Hit": (
                    "The percentage of how many lookups there are from Adapative Hash Index compared to it not"
                    " being used"
                ),
                "Diff": "This is the size difference of the binary log between each refresh interval",
                "Cache Hit": "The percentage of how many binary log lookups are from cache instead of from disk",
                "History List": "History list length (number of un-purged row changes in InnoDB's undo logs)",
                "QPS": "Queries per second from Com_queries in SHOW GLOBAL STATUS",
                "Latency": "How much time it takes to receive data from the host for each refresh interval",
                "Threads": "Con = Connected, Run = Running, Cac = Cached from SHOW GLOBAL STATUS",
                "Speed": "How many seconds were taken off of replication lag from the last refresh interval",
                "Tickets": "Relates to innodb_concurrency_tickets variable",
                "R-Lock/Mod": "Relates to how many rows are locked/modified for the thread's transaction",
                "GR": "Group Replication",
            }

            table_terminology = Table(
                box=box.SIMPLE_HEAVY,
                show_edge=False,
                style="table_border",
                title="Terminology",
                title_style="bold #bbc8e8",
                header_style="bold",
            )
            table_terminology.add_column("Datapoint", style="highlight")
            table_terminology.add_column("Description")
            for datapoint, description in sorted(datapoints.items()):
                table_terminology.add_row(datapoint, description)

            screen_data = Group(
                Align.center(table_keys),
                "",
                Align.center(table_terminology),
                "",
                Align.center(
                    "[light_blue][b]Note[/b]: Textual puts your terminal in application mode which disables selecting"
                    " text.\nTo see how to select text on your terminal, visit: https://tinyurl.com/dolphie-copy-text"
                ),
            )

        if screen_data:
            self.app.push_screen(
                CommandScreen(dolphie.read_only_status, dolphie.app_version, dolphie.mysql_host, screen_data)
            )

    @work(thread=True)
    def run_command_in_worker(self, key: str, dolphie: Dolphie, additional_data=None):
        tab = self.tab_manager.get_tab(self.tab.id)

        # These are the screens to display we use for the commands
        def show_command_screen():
            self.app.push_screen(
                CommandScreen(dolphie.read_only_status, dolphie.app_version, dolphie.mysql_host, screen_data)
            )

        def show_thread_screen():
            self.app.push_screen(
                ThreadScreen(
                    read_only=dolphie.read_only_status,
                    app_version=dolphie.app_version,
                    host=dolphie.mysql_host,
                    thread_table=thread_table,
                    user_thread_attributes_table=user_thread_attributes_table,
                    query=formatted_query,
                    explain_data=explain_data,
                    explain_failure=explain_failure,
                    transaction_history_table=transaction_history_table,
                )
            )

        tab.spinner.show()

        try:
            if key == "d":
                tables = {}
                all_tables = []

                db_count = dolphie.secondary_db_connection.execute(MySQLQueries.databases)
                databases = dolphie.secondary_db_connection.fetchall()

                # Determine how many tables to provide data
                max_num_tables = 1 if db_count <= 20 else 3

                # Calculate how many databases per table
                row_per_count = db_count // max_num_tables

                # Create dictionary of tables
                for table_counter in range(1, max_num_tables + 1):
                    table_box = box.HORIZONTALS
                    if max_num_tables == 1:
                        table_box = None

                    tables[table_counter] = Table(box=table_box, show_header=False, style="table_border")
                    tables[table_counter].add_column("")

                # Loop over databases
                db_counter = 1
                table_counter = 1

                # Sort the databases by name
                for database in databases:
                    tables[table_counter].add_row(database["SCHEMA_NAME"])
                    db_counter += 1

                    if db_counter > row_per_count and table_counter < max_num_tables:
                        table_counter += 1
                        db_counter = 1

                # Collect table data into an array
                all_tables = [table_data for table_data in tables.values() if table_data]

                table_grid = Table.grid()
                table_grid.add_row(*all_tables)

                screen_data = Group(
                    Align.center("[b light_blue]Databases[/b light_blue] ([b highlight]%s[/b highlight])\n" % db_count),
                    Align.center(table_grid),
                )

                self.call_from_thread(show_command_screen)

            if key == "k":
                thread_id = additional_data
                try:
                    if dolphie.aws_rds:
                        dolphie.secondary_db_connection.execute("CALL mysql.rds_kill(%s)" % thread_id)
                    else:
                        dolphie.secondary_db_connection.execute("KILL %s" % thread_id)

                    self.notify("Killed Thread ID [b highlight]%s[/b highlight]" % thread_id, severity="success")
                except ManualException as e:
                    self.notify(e.reason, title="Error killing Thread ID", severity="error")

            if key == "K":

                def execute_kill(thread_id):
                    query = "CALL mysql.rds_kill(%s)" if dolphie.aws_rds else "KILL %s"
                    dolphie.secondary_db_connection.execute(query % thread_id)

                kill_type, kill_value, include_sleeping_queries, lower_limit, upper_limit = additional_data
                db_field = {"username": "user", "host": "host", "time_range": "time"}.get(kill_type)

                commands_to_kill = ["Query", "Execute"]

                if include_sleeping_queries:
                    commands_to_kill.append("Sleep")

                tab.spinner.show()
                threads_killed = 0
                for thread_id, thread in dolphie.processlist_threads_snapshot.items():
                    thread: ProcesslistThread
                    try:
                        if thread.command in commands_to_kill:
                            if kill_type == "time_range":
                                if lower_limit <= thread.time <= upper_limit:
                                    execute_kill(thread_id)
                                    threads_killed += 1
                            elif getattr(thread, db_field) == kill_value:
                                execute_kill(thread_id)
                                threads_killed += 1
                    except ManualException as e:
                        self.notify(e.reason, title=f"Error Killing Thread ID {thread_id}", severity="error")

                if threads_killed:
                    self.notify(f"Killed [highlight]{threads_killed}[/highlight] threads")
                else:
                    self.notify("No threads were killed")

            elif key == "l":
                deadlock = ""
                output = re.search(
                    r"------------------------\nLATEST\sDETECTED\sDEADLOCK\n------------------------"
                    "\n(.*?)------------\nTRANSACTIONS",
                    dolphie.secondary_db_connection.fetch_value_from_field(MySQLQueries.innodb_status, "Status"),
                    flags=re.S,
                )
                if output:
                    deadlock = output.group(1)

                    deadlock = deadlock.replace("***", "[yellow]*****[/yellow]")
                    screen_data = deadlock
                else:
                    screen_data = Align.center("No deadlock detected")

                self.call_from_thread(show_command_screen)

            elif key == "o":
                screen_data = dolphie.secondary_db_connection.fetch_value_from_field(
                    MySQLQueries.innodb_status, "Status"
                )

                self.call_from_thread(show_command_screen)

            elif key == "m":
                table_grid = Table.grid()
                table1 = Table(box=box.SIMPLE_HEAVY, style="table_border")

                header_style = Style(bold=True)
                table1.add_column("User", header_style=header_style)
                table1.add_column("Current", header_style=header_style)
                table1.add_column("Total", header_style=header_style)

                dolphie.secondary_db_connection.execute(MySQLQueries.memory_by_user)
                data = dolphie.secondary_db_connection.fetchall()
                for row in data:
                    table1.add_row(
                        row["user"],
                        format_sys_table_memory(row["current_allocated"]),
                        format_sys_table_memory(row["total_allocated"]),
                    )

                table2 = Table(box=box.SIMPLE_HEAVY, style="table_border")
                table2.add_column("Code Area", header_style=header_style)
                table2.add_column("Current", header_style=header_style)

                dolphie.secondary_db_connection.execute(MySQLQueries.memory_by_code_area)
                data = dolphie.secondary_db_connection.fetchall()
                for row in data:
                    table2.add_row(row["code_area"], format_sys_table_memory(row["current_allocated"]))

                table3 = Table(box=box.SIMPLE_HEAVY, style="table_border")
                table3.add_column("Host", header_style=header_style)
                table3.add_column("Current", header_style=header_style)
                table3.add_column("Total", header_style=header_style)

                dolphie.secondary_db_connection.execute(MySQLQueries.memory_by_host)
                data = dolphie.secondary_db_connection.fetchall()
                for row in data:
                    table3.add_row(
                        dolphie.get_hostname(row["host"]),
                        format_sys_table_memory(row["current_allocated"]),
                        format_sys_table_memory(row["total_allocated"]),
                    )

                table_grid.add_row("", Align.center("[b light_blue]Memory Allocation"), "")
                table_grid.add_row(table1, table3, table2)

                screen_data = Align.center(table_grid)

                self.call_from_thread(show_command_screen)
            elif key == "t":
                formatted_query = ""
                explain_failure = ""
                explain_data = ""

                thread_table = Table(box=None, show_header=False)
                thread_table.add_column("")
                thread_table.add_column("", overflow="fold")

                thread_id = additional_data
                thread_data: ProcesslistThread = dolphie.processlist_threads_snapshot.get(thread_id)
                thread_table.add_row("[label]Thread ID", thread_id)
                thread_table.add_row("[label]User", thread_data.user)
                thread_table.add_row("[label]Host", thread_data.host)
                thread_table.add_row("[label]Database", thread_data.db)
                thread_table.add_row("[label]Command", thread_data.command)
                thread_table.add_row("[label]State", thread_data.state)
                thread_table.add_row("[label]Time", str(timedelta(seconds=thread_data.time)).zfill(8))
                thread_table.add_row("[label]Rows Locked", thread_data.trx_rows_locked)
                thread_table.add_row("[label]Rows Modified", thread_data.trx_rows_modified)

                if dolphie.global_variables.get("innodb_thread_concurrency"):
                    thread_table.add_row("[label]Tickets", thread_data.trx_concurrency_tickets)

                thread_table.add_row("", "")
                thread_table.add_row("[label]TRX Time", thread_data.trx_time)
                thread_table.add_row("[label]TRX State", thread_data.trx_state)
                thread_table.add_row("[label]TRX Operation", thread_data.trx_operation_state)

                query = sqlformat(thread_data.query, reindent_aligned=True)
                query_db = thread_data.db
                if query:
                    formatted_query = Syntax(
                        query,
                        "sql",
                        line_numbers=False,
                        word_wrap=True,
                        theme="monokai",
                        background_color="#101626",
                    )

                    if query_db:
                        try:
                            dolphie.secondary_db_connection.execute("USE %s" % query_db)
                            dolphie.secondary_db_connection.execute("EXPLAIN %s" % query)

                            explain_data = dolphie.secondary_db_connection.fetchall()
                        except ManualException as e:
                            explain_failure = "[b indian_red]EXPLAIN ERROR:[/b indian_red] [indian_red]%s" % e.reason

                user_thread_attributes_table = None
                if dolphie.performance_schema_enabled:
                    user_thread_attributes_table = Table(box=None, show_header=False, expand=True)

                    dolphie.secondary_db_connection.execute(
                        MySQLQueries.user_thread_attributes.replace("$1", thread_id)
                    )

                    user_thread_attributes = dolphie.secondary_db_connection.fetchall()
                    if user_thread_attributes:
                        user_thread_attributes_table.add_column("")
                        user_thread_attributes_table.add_column("", overflow="fold")

                        for attribute in user_thread_attributes:
                            user_thread_attributes_table.add_row(
                                f"[label]{attribute['ATTR_NAME']}", attribute["ATTR_VALUE"]
                            )
                    else:
                        user_thread_attributes_table.add_column(justify="center")
                        user_thread_attributes_table.add_row("[b][label]None found")

                # Transaction history
                transaction_history_table = None
                if (
                    dolphie.is_mysql_version_at_least("5.7")
                    and dolphie.performance_schema_enabled
                    and thread_data.mysql_thread_id
                ):
                    query = MySQLQueries.thread_transaction_history.replace("$1", str(thread_data.mysql_thread_id))
                    dolphie.secondary_db_connection.execute(query)
                    transaction_history = dolphie.secondary_db_connection.fetchall()

                    if transaction_history:
                        transaction_history_table = Table(box=None)
                        transaction_history_table.add_column("Start Time")
                        transaction_history_table.add_column("Query", overflow="fold")

                        for query in transaction_history:
                            trx_history_formatted_query = ""
                            if query["sql_text"]:
                                trx_history_formatted_query = Syntax(
                                    re.sub(r"\s+", " ", query["sql_text"]),
                                    "sql",
                                    line_numbers=False,
                                    word_wrap=True,
                                    theme="monokai",
                                    background_color="#101626",
                                )

                            transaction_history_table.add_row(
                                query["start_time"].strftime("%Y-%m-%d %H:%M:%S"), trx_history_formatted_query
                            )

                self.call_from_thread(show_thread_screen)

            elif key == "u":
                if dolphie.is_mysql_version_at_least("5.7"):
                    dolphie.secondary_db_connection.execute(MySQLQueries.ps_user_statisitics)
                else:
                    dolphie.secondary_db_connection.execute(MySQLQueries.ps_user_statisitics_56)

                users = dolphie.secondary_db_connection.fetchall()

                columns = {
                    "User": {"field": "user", "format_number": False},
                    "Active": {"field": "current_connections", "format_number": True},
                    "Total": {"field": "total_connections", "format_number": True},
                    "Rows Read": {"field": "rows_examined", "format_number": True},
                    "Rows Sent": {"field": "rows_sent", "format_number": True},
                    "Rows Updated": {"field": "rows_affected", "format_number": True},
                    "Tmp Tables": {"field": "created_tmp_tables", "format_number": True},
                    "Tmp Disk Tables": {"field": "created_tmp_disk_tables", "format_number": True},
                    "Plugin": {"field": "plugin", "format_number": False},
                    "Password Expire": {"field": "password_expires_in", "format_number": False},
                }

                table = Table(
                    header_style="b",
                    box=box.SIMPLE_HEAVY,
                    show_edge=False,
                    style="table_border",
                )
                for column, data in columns.items():
                    table.add_column(column, no_wrap=True)

                for user in users:
                    row_values = []

                    for column, data in columns.items():
                        value = user.get(data["field"], "N/A")

                        if data["format_number"]:
                            row_values.append(format_number(value) if value else "0")
                        else:
                            row_values.append(value or "")

                    table.add_row(*row_values)

                screen_data = Group(
                    Align.center(f"[b light_blue]Users Connected ([highlight]{len(users)}[/highlight])\n"),
                    Align.center(table),
                )

                self.call_from_thread(show_command_screen)
        except ManualException as e:
            self.notify(e.reason, title=f"Error running command '{key}'", severity="error", timeout=10)

        tab.spinner.hide()

    def compose(self) -> ComposeResult:
        yield TopBar(host="", app_version=self.dolphie.app_version, help="press [b highlight]?[/b highlight] for help")
        yield TabbedContent(id="tabbed_content")


def main():
    # Set environment variables so Textual can use all the pretty colors
    os.environ["TERM"] = "xterm-256color"
    os.environ["COLORTERM"] = "truecolor"

    dolphie = Dolphie()
    parse_args(dolphie)

    app = DolphieApp(dolphie)
    app.run()


if __name__ == "__main__":
    main()



import copy
from dataclasses import dataclass

import dolphie.Modules.MetricManager as MetricManager
from dolphie import Dolphie
from dolphie.Widgets.host_setup import HostSetupModal
from dolphie.Widgets.spinner import SpinnerWidget
from dolphie.Widgets.topbar import TopBar
from textual.app import App
from textual.containers import (
    Center,
    Container,
    Horizontal,
    ScrollableContainer,
    VerticalScroll,
)
from textual.timer import Timer
from textual.widgets import (
    DataTable,
    Label,
    LoadingIndicator,
    Sparkline,
    Static,
    Switch,
    TabbedContent,
    TabPane,
)
from textual.worker import Worker, WorkerState


@dataclass
class Tab:
    id: int
    name: str
    dolphie: Dolphie

    worker: Worker = None
    worker_timer: Timer = None
    worker_cancel_error: str = None

    replicas_worker: worker = None
    replicas_worker_timer: Timer = None

    topbar: TopBar = None
    main_container: VerticalScroll = None
    loading_indicator: LoadingIndicator = None
    sparkline: Sparkline = None
    panel_dashboard: Container = None
    panel_graphs: Container = None
    panel_replication: Container = None
    panel_locks: Container = None
    panel_ddl: Container = None
    panel_processlist: Container = None

    spinner: SpinnerWidget = None

    dashboard_host_information: Static = None
    dashboard_innodb: Static = None
    dashboard_binary_log: Static = None
    dashboard_statistics: Static = None
    dashboard_replication: Static = None

    ddl_title: Label = None
    ddl_datatable: DataTable = None

    locks_title: Label = None
    locks_datatable: DataTable = None

    processlist_title: Label = None
    processlist_datatable: DataTable = None

    replication_container_title: Static = None
    replication_container: Container = None
    replication_variables: Label = None
    replication_status: Static = None
    replication_thread_applier_container: ScrollableContainer = None
    replication_thread_applier: Static = None

    group_replication_container: Container = None
    group_replication_grid: Container = None
    group_replication_title: Label = None
    group_replication_data: Static = None

    replicas_container: Container = None
    replicas_grid: Container = None
    replicas_loading_indicator: LoadingIndicator = None
    replicas_title: Label = None

    cluster_data: Static = None

    queue_for_removal: bool = False

    def disconnect(self, update_topbar: bool = True):
        if self.worker_timer:
            self.worker_timer.stop()
        if self.replicas_worker_timer:
            self.replicas_worker_timer.stop()

        if self.worker:
            self.worker.cancel()
        self.worker = None

        if self.replicas_worker:
            self.replicas_worker.cancel()
        self.replicas_worker = None

        if self.dolphie.main_db_connection:
            self.dolphie.main_db_connection.close()

        if self.dolphie.secondary_db_connection:
            self.dolphie.secondary_db_connection.close()

        self.dolphie.replica_manager.remove_all()

        self.main_container.display = False
        self.sparkline.display = False

        self.replicas_title.update("")
        for member in self.dolphie.app.query(f".replica_container_{self.id}"):
            member.remove()

        if update_topbar:
            self.update_topbar()

    def update_topbar(self, custom_text: str = None):
        dolphie = self.dolphie

        if custom_text:
            self.topbar.host = custom_text
            return

        if dolphie.main_db_connection and not dolphie.main_db_connection.is_connected():
            dolphie.read_only_status = "DISCONNECTED"
        else:
            if not self.worker:
                if not self.loading_indicator.display:
                    self.topbar.host = ""

                # If there is no worker instance, we don't update the topbar
                return

        if dolphie.read_only_status and dolphie.mysql_host:
            self.topbar.host = f"[[white]{dolphie.read_only_status}[/white]] {dolphie.mysql_host}"

    def host_setup(self):
        dolphie = self.dolphie

        def command_get_input(data):
            host_port = data["host"].split(":")

            dolphie.host = host_port[0]
            dolphie.port = int(host_port[1]) if len(host_port) > 1 else 3306

            password = data.get("password")
            if password:
                dolphie.password = password

            self.disconnect()
            dolphie.reset_runtime_variables()

            if not self.worker or self.worker.state == WorkerState.CANCELLED:
                self.worker_cancel_error = ""

                self.dolphie.app.run_worker_main(self.id)
                self.dolphie.app.run_worker_replicas(self.id)

        self.loading_indicator.display = False

        # If we're here because of a worker cancel error, we want to pre-populate the host/port
        if self.worker_cancel_error:
            host = dolphie.host
            port = dolphie.port

            self.update_topbar()
        else:
            host = ""
            port = ""

        self.dolphie.app.push_screen(
            HostSetupModal(
                host=host,
                port=port,
                available_hosts=dolphie.host_setup_available_hosts,
                error_message=self.worker_cancel_error,
            ),
            command_get_input,
        )


class TabManager:
    def __init__(self, app: App):
        self.app = app

        self.tabs: dict = {}
        self.tab_id_counter: int = 1

        self.tabbed_content = self.app.query_one("#tabbed_content", TabbedContent)

    async def create_tab(self, tab_name: str, dolphie: Dolphie):
        tab_id = self.tab_id_counter

        await self.tabbed_content.add_pane(
            TabPane(
                tab_name,
                LoadingIndicator(id=f"loading_indicator_{tab_id}"),
                SpinnerWidget(id=f"spinner_{tab_id}"),
                VerticalScroll(
                    Container(
                        Center(
                            Static(id=f"dashboard_host_information_{tab_id}", classes="dashboard_host_information"),
                            Static(id=f"dashboard_innodb_{tab_id}", classes="dashboard_innodb_information"),
                            Static(id=f"dashboard_binary_log_{tab_id}", classes="dashboard_binary_log"),
                            Static(id=f"dashboard_replication_{tab_id}", classes="dashboard_replication"),
                            Static(id=f"dashboard_statistics_{tab_id}", classes="dashboard_statistics"),
                        ),
                        Sparkline([], id=f"panel_dashboard_queries_qps_{tab_id}"),
                        id=f"panel_dashboard_{tab_id}",
                        classes="panel_container dashboard",
                    ),
                    Container(
                        TabbedContent(id=f"tabbed_content_{tab_id}", classes="metrics_tabbed_content"),
                        id=f"panel_graphs_{tab_id}",
                        classes="panel_container",
                    ),
                    Container(
                        Static(id=f"replication_container_title_{tab_id}", classes="replication_container_title"),
                        Container(
                            Label("[b]Replication\n"),
                            Label(id=f"replication_variables_{tab_id}"),
                            Center(
                                ScrollableContainer(
                                    Static(id=f"replication_status_{tab_id}"), classes="replication_status"
                                ),
                                ScrollableContainer(
                                    Static(id=f"replication_thread_applier_{tab_id}"),
                                    id=f"replication_thread_applier_container_{tab_id}",
                                    classes="replication_thread_applier",
                                ),
                            ),
                            id=f"replication_container_{tab_id}",
                            classes="replication",
                        ),
                        Container(
                            Label(id=f"group_replication_title_{tab_id}"),
                            Label(id=f"group_replication_data_{tab_id}"),
                            Container(id=f"group_replication_grid_{tab_id}"),
                            id=f"group_replication_container_{tab_id}",
                            classes="group_replication",
                        ),
                        Static(id=f"cluster_data_{tab_id}"),
                        Container(
                            Label(id=f"replicas_title_{tab_id}"),
                            LoadingIndicator(id=f"replicas_loading_indicator_{tab_id}"),
                            Container(id=f"replicas_grid_{tab_id}"),
                            id=f"replicas_container_{tab_id}",
                            classes="replicas",
                        ),
                        id=f"panel_replication_{tab_id}",
                        classes="panel_container replication_panel",
                    ),
                    Container(
                        Label(id=f"locks_title_{tab_id}"),
                        DataTable(id=f"locks_datatable_{tab_id}", show_cursor=False),
                        id=f"panel_locks_{tab_id}",
                        classes="locks",
                    ),
                    Container(
                        Label(id=f"ddl_title_{tab_id}"),
                        DataTable(id=f"ddl_datatable_{tab_id}", show_cursor=False),
                        id=f"panel_ddl_{tab_id}",
                        classes="ddl",
                    ),
                    Container(
                        Label(id=f"processlist_title_{tab_id}"),
                        DataTable(id=f"processlist_data_{tab_id}", show_cursor=False),
                        id=f"panel_processlist_{tab_id}",
                        classes="processlist",
                    ),
                    classes="tab",
                    id=f"main_container_{tab_id}",
                ),
                id=f"tab_{tab_id}",
                name=tab_id,
            ),
        )

        metrics = dolphie.metric_manager.metrics
        metric_tab_labels = [
            ("DML", metrics.dml, True),
            ("Locks", metrics.locks, False),
            ("Table Cache", metrics.table_cache, True),
            ("Threads", metrics.threads, True),
            ("BP Requests", metrics.buffer_pool_requests, True),
            ("Checkpoint", metrics.checkpoint, False),
            ("Redo Log", metrics.redo_log, False),
            ("AHI", metrics.adaptive_hash_index, True),
            ("Temp Objects", metrics.temporary_objects, True),
            ("Aborted Connections", metrics.aborted_connections, True),
            ("Disk I/O", metrics.disk_io, True),
            ("Replication", metrics.replication_lag, False),
        ]

        for tab_formatted_name, metric_instance, create_switches in metric_tab_labels:
            metric_tab_name = metric_instance.tab_name
            graph_names = metric_instance.graphs

            await self.app.query_one(f"#tabbed_content_{tab_id}", TabbedContent).add_pane(
                TabPane(
                    tab_formatted_name,
                    Label(id=f"stats_{metric_tab_name}_{tab_id}", classes="stats_data"),
                    Horizontal(id=f"graph_container_{metric_tab_name}_{tab_id}"),
                    Horizontal(
                        id=f"switch_container_{metric_tab_name}_{tab_id}",
                        classes=f"switch_container_{tab_id} switch_container",
                    ),
                    id=f"graph_tab_{metric_tab_name}_{tab_id}",
                    name=metric_tab_name,
                )
            )

            if metric_tab_name == "redo_log":
                graph_names = ["graph_redo_log", "graph_redo_log_active_count", "graph_redo_log_bar"]
            elif metric_tab_name == "adaptive_hash_index":
                graph_names = ["graph_adaptive_hash_index", "graph_adaptive_hash_index_hit_ratio"]

            for graph_name in graph_names:
                await self.app.query_one(f"#graph_container_{metric_tab_name}_{tab_id}", Horizontal).mount(
                    MetricManager.Graph(id=f"{graph_name}_{tab_id}", classes="panel_data")
                )

            if create_switches:
                for metric, metric_data in metric_instance.__dict__.items():
                    if isinstance(metric_data, MetricManager.MetricData) and metric_data.graphable:
                        await self.app.query_one(f"#switch_container_{metric_tab_name}_{tab_id}", Horizontal).mount(
                            Label(metric_data.label)
                        )
                        await self.app.query_one(f"#switch_container_{metric_tab_name}_{tab_id}", Horizontal).mount(
                            Switch(animate=False, id=metric, name=metric_tab_name)
                        )

        # Create a new tab instance
        dolphie = copy.copy(dolphie)
        dolphie.reset_runtime_variables()
        if tab_id != 1:
            dolphie.host = ""
            dolphie.port = ""

        tab = Tab(id=tab_id, name=tab_name, dolphie=dolphie)

        dolphie.tab_id = tab_id
        dolphie.tab_name = tab_name

        # Save references to the widgets in the tab
        tab.topbar = self.app.query_one(TopBar)
        tab.main_container = self.app.query_one(f"#main_container_{tab.id}", VerticalScroll)
        tab.loading_indicator = self.app.query_one(f"#loading_indicator_{tab.id}", LoadingIndicator)
        tab.sparkline = self.app.query_one(f"#panel_dashboard_queries_qps_{tab.id}", Sparkline)
        tab.panel_dashboard = self.app.query_one(f"#panel_dashboard_{tab.id}", Container)
        tab.panel_graphs = self.app.query_one(f"#panel_graphs_{tab.id}", Container)
        tab.panel_replication = self.app.query_one(f"#panel_replication_{tab.id}", Container)
        tab.panel_locks = self.app.query_one(f"#panel_locks_{tab.id}", Container)
        tab.panel_processlist = self.app.query_one(f"#panel_processlist_{tab.id}", Container)
        tab.panel_ddl = self.app.query_one(f"#panel_ddl_{tab.id}", Container)

        tab.spinner = self.app.query_one(f"#spinner_{tab.id}", SpinnerWidget)
        tab.spinner.hide()

        tab.ddl_title = self.app.query_one(f"#ddl_title_{tab.id}", Label)
        tab.ddl_datatable = self.app.query_one(f"#ddl_datatable_{tab.id}", DataTable)
        tab.processlist_title = self.app.query_one(f"#processlist_title_{tab.id}", Label)
        tab.processlist_datatable = self.app.query_one(f"#processlist_data_{tab.id}", DataTable)
        tab.locks_title = self.app.query_one(f"#locks_title_{tab.id}", Label)
        tab.locks_datatable = self.app.query_one(f"#locks_datatable_{tab.id}", DataTable)

        tab.dashboard_host_information = self.app.query_one(f"#dashboard_host_information_{tab.id}", Static)
        tab.dashboard_innodb = self.app.query_one(f"#dashboard_innodb_{tab.id}", Static)
        tab.dashboard_binary_log = self.app.query_one(f"#dashboard_binary_log_{tab.id}", Static)
        tab.dashboard_statistics = self.app.query_one(f"#dashboard_statistics_{tab.id}", Static)
        tab.dashboard_replication = self.app.query_one(f"#dashboard_replication_{tab.id}", Static)

        tab.group_replication_container = self.app.query_one(f"#group_replication_container_{tab.id}", Container)
        tab.group_replication_grid = self.app.query_one(f"#group_replication_grid_{tab.id}", Container)

        tab.replicas_grid = self.app.query_one(f"#replicas_grid_{tab.id}", Container)
        tab.replicas_container = self.app.query_one(f"#replicas_container_{tab.id}", Container)

        tab.group_replication_data = self.app.query_one(f"#group_replication_data_{tab.id}", Static)
        tab.group_replication_title = self.app.query_one(f"#group_replication_title_{tab.id}", Label)
        tab.replicas_title = self.app.query_one(f"#replicas_title_{tab.id}", Label)
        tab.replicas_loading_indicator = self.app.query_one(f"#replicas_loading_indicator_{tab.id}", LoadingIndicator)

        tab.cluster_data = self.app.query_one(f"#cluster_data_{tab.id}", Static)

        tab.replication_container_title = self.app.query_one(f"#replication_container_title_{tab.id}", Static)
        tab.replication_container = self.app.query_one(f"#replication_container_{tab.id}", Container)
        tab.replication_variables = self.app.query_one(f"#replication_variables_{tab.id}", Label)
        tab.replication_status = self.app.query_one(f"#replication_status_{tab.id}", Static)
        tab.replication_thread_applier_container = self.app.query_one(
            f"#replication_thread_applier_container_{tab.id}", ScrollableContainer
        )
        tab.replication_thread_applier = self.app.query_one(f"#replication_thread_applier_{tab.id}", Static)

        # By default, hide all the panels
        tab.sparkline.display = False
        for panel in tab.dolphie.panels.all():
            self.app.query_one(f"#panel_{panel}_{tab.id}").display = False

        # Set panels to be visible for the ones the user specifies
        for panel in dolphie.startup_panels:
            setattr(getattr(dolphie.panels, panel), "visible", True)

        # Set what marker we use for graphs
        graphs = self.app.query(MetricManager.Graph)
        for graph in graphs:
            graph.marker = dolphie.graph_marker

        # Set default switches to be toggled on
        switches = self.app.query(f".switch_container_{tab_id} Switch")
        switches_to_toggle = [switch for switch in switches if switch.id not in ["Queries", "Threads_connected"]]
        for switch in switches_to_toggle:
            switch.toggle()

        # Save the tab instance to the tab manager
        self.tabs[tab_id] = tab

        # Switch to the new tab
        self.switch_tab(tab_id)

        # Increment the tab id counter
        self.tab_id_counter += 1

    async def remove_tab(self, tab_id: int):
        await self.tabbed_content.remove_pane(f"tab_{self.get_tab(tab_id).id}")

        tab = self.get_tab(tab_id)
        tab.queue_for_removal = True

    def switch_tab(self, tab_id: int):
        tab = self.get_tab(tab_id)

        self.app.tab = tab  # Update the current tab variable for the app

        self.tabbed_content.active = f"tab_{tab.id}"

        tab.update_topbar()

    def get_tab(self, id: int) -> Tab:
        if id in self.tabs:
            return self.tabs[id]

    def get_all_tabs(self) -> list:
        all_tabs = []

        for tab in self.tabs.values():
            tab: Tab
            if not tab.queue_for_removal:
                all_tabs.append(tab.id)

        return all_tabs



from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Input, Label, Static
from textual_autocomplete import AutoComplete, Dropdown, DropdownItem


class HostSetupModal(ModalScreen):
    CSS = """
        HostSetupModal > Vertical {
            background: #131626;
            border: tall #384673;
            height: auto;
            width: 70;
        }
        HostSetupModal > Vertical > * {
            width: auto;
            height: auto;
            align: center middle;
        }
        HostSetupModal Label {
            text-style: bold;
            width: 100%;
            content-align: center middle;
            padding-bottom: 1;
        }
        HostSetupModal Input {
            width: 100% !important;
            content-align: center middle;
        }
        HostSetupModal .main_container {
            width: 100%;
            content-align: center middle;
        }
        HostSetupModal AutoComplete {
            width: 100%;
            height: auto;
        }
        HostSetupModal #modal_footer {
            color: #d3565c;
            width: 100%;
            padding-bottom: 0;
            padding-top: 1;
            margin: 0 2;
        }
    """
    BINDINGS = [
        Binding("escape", "app.pop_screen", "", show=False),
    ]

    def __init__(self, host, port, available_hosts, error_message=None):
        super().__init__()

        self.host = host
        self.port = port

        if self.host and self.port:
            self.host = f"{self.host}:{self.port}"

        self.dropdown_items = []
        if available_hosts:
            self.dropdown_items = [DropdownItem(id) for id in sorted(available_hosts)]

        self.error_message = error_message

    def on_mount(self) -> None:
        footer = self.query_one("#modal_footer", Static)

        footer.display = False
        if self.error_message:
            footer.update(self.error_message)
            footer.display = True

    def compose(self) -> ComposeResult:
        with Vertical():
            with Vertical(classes="main_container"):
                yield Label("Host Setup")
                yield AutoComplete(
                    Input(
                        value=self.host,
                        id="host",
                        placeholder="Start typing to search for a host - format is host:port",
                    ),
                    Dropdown(id="dropdown_items", items=self.dropdown_items),
                )
                yield Input(id="password", placeholder="Password (empty for current)", password=True)
            with Horizontal(classes="button_container"):
                yield Button("Submit", id="submit", variant="primary")
                yield Button("Cancel", id="cancel")
            yield Label(id="modal_footer")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "submit":
            error_message = None

            host = self.query_one("#host", Input)
            password = self.query_one("#password", Input)

            modal_footer = self.query_one("#modal_footer", Label)
            host_split = host.value.split(":")
            if len(host_split) == 2:
                try:
                    int(host_split[1])
                except ValueError:
                    error_message = "Port must be a valid integer"
            elif len(host_split) > 2:
                error_message = "Host cannot contain more than one colon"

            if not host.value:
                error_message = "Host cannot be empty"

            if error_message:
                modal_footer.update(error_message)
                modal_footer.display = True
                return

            self.dismiss({"host": host.value, "password": password.value})
        else:
            self.app.pop_screen()

    @on(Input.Submitted, "#password")
    def on_input_submitted(self):
        self.query_one("#submit", Button).press()



from typing import Dict

from dolphie.DataTypes import ProcesslistThread
from dolphie.Modules.Functions import format_number
from dolphie.Modules.Queries import MySQLQueries
from dolphie.Modules.TabManager import Tab
from textual.widgets import DataTable


def create_panel(tab: Tab) -> DataTable:
    dolphie = tab.dolphie

    if not dolphie.performance_schema_enabled and dolphie.use_performance_schema:
        dolphie.notify("Performance Schema is not enabled on this host, using Information Schema instead")
        dolphie.use_performance_schema = False

    columns = [
        {"name": "Thread ID", "field": "id", "width": 11, "format_number": False},
        {"name": "Username", "field": "user", "width": 13, "format_number": False},
    ]

    if dolphie.show_additional_query_columns:
        columns.extend([
            {"name": "Hostname/IP", "field": "host", "width": 16, "format_number": False},
            {"name": "Database", "field": "db", "width": 13, "format_number": False},
        ])

    columns.extend([
        {"name": "Command", "field": "command", "width": 8, "format_number": False},
        {"name": "State", "field": "state", "width": 16, "format_number": False},
        {"name": "TRX State", "field": "trx_state", "width": 9, "format_number": False},
        {"name": "R-Lock", "field": "trx_rows_locked", "width": 7, "format_number": True},
        {"name": "R-Mod", "field": "trx_rows_modified", "width": 7, "format_number": True},
    ])

    if dolphie.show_additional_query_columns and dolphie.global_variables.get("innodb_thread_concurrency"):
        columns.append({"name": "Tickets", "field": "trx_concurrency_tickets", "width": 8, "format_number": True})

    if dolphie.show_trxs_only:
        columns.append(
            {"name": "TRX Time", "field": "trx_time", "width": 9, "format_number": False},
        )

    columns.extend([
        {"name": "Time", "field": "formatted_time", "width": 9, "format_number": False},
        {"name": "Query", "field": "formatted_query", "width": None, "format_number": False},
        {"name": "time_seconds", "field": "time", "width": 0, "format_number": False},
    ])

    processlist_datatable = tab.processlist_datatable

    # Clear table if columns change
    if len(processlist_datatable.columns) != len(columns):
        processlist_datatable.clear(columns=True)

    # Add columns to the datatable if it is empty
    if not processlist_datatable.columns:
        for column_data in columns:
            column_name = column_data["name"]
            column_width = column_data["width"]
            processlist_datatable.add_column(column_name, key=column_name, width=column_width)

    # Iterate through dolphie.processlist_threads
    for thread_id, thread in dolphie.processlist_threads.items():
        # Check if the thread_id exists in the datatable
        if thread_id in processlist_datatable.rows:
            datatable_row = processlist_datatable.get_row(thread_id)

            # Update the datatable if values differ
            for column_id, column_data in enumerate(columns):
                column_name = column_data["name"]
                column_field = column_data["field"]
                column_format_number = column_data["format_number"]
                column_value = getattr(thread, column_field)

                update_width = False
                if column_field == "formatted_query":
                    update_width = True

                value = format_number(column_value) if column_format_number else column_value
                if value != datatable_row[column_id] or column_field == "formatted_time":
                    processlist_datatable.update_cell(thread_id, column_name, value, update_width=update_width)
        else:
            # Add a new row to the datatable if thread_id does not exist
            row_values = []
            for column_data in columns:
                column_field = column_data["field"]
                column_format_number = column_data["format_number"]
                column_value = getattr(thread, column_field)

                value = format_number(column_value) if column_format_number else column_value

                row_values.append(value)

            processlist_datatable.add_row(*row_values, key=thread_id)

    # Remove rows from processlist_datatable that no longer exist in dolphie.processlist_threads
    rows_to_remove = set(processlist_datatable.rows.keys()) - set(dolphie.processlist_threads.keys())
    for id in rows_to_remove:
        processlist_datatable.remove_row(id)

    processlist_datatable.sort("time_seconds", reverse=dolphie.sort_by_time_descending)

    tab.processlist_title.update(f"Processlist ([highlight]{len(dolphie.processlist_threads)}[/highlight])")


def fetch_data(tab: Tab) -> Dict[str, ProcesslistThread]:
    dolphie = tab.dolphie

    if dolphie.use_performance_schema:
        processlist_query = MySQLQueries.ps_query
    else:
        processlist_query = MySQLQueries.pl_query

    ########################
    # WHERE clause filters #
    ########################
    where_clause = []

    # Filter out idle threads if specified
    if not dolphie.show_idle_threads:
        if dolphie.use_performance_schema:
            where_clause.append(
                "(processlist_command != 'Sleep' AND processlist_command NOT LIKE 'Binlog Dump%') AND (processlist_info"
                " IS NOT NULL OR trx_query IS NOT NULL) AND IFNULL(processlist_state, '') NOT LIKE 'Group Replication"
                " Module%'"
            )
        else:
            where_clause.append(
                "(Command != 'Sleep' AND Command NOT LIKE 'Binlog Dump%') AND (Info IS NOT NULL OR trx_query IS NOT"
                " NULL) AND IFNULL(State, '') NOT LIKE 'Group Replication Module%'"
            )

    # Only show running transactions only
    if dolphie.show_trxs_only:
        where_clause.append("trx_state != ''")

    # Filter user
    if dolphie.user_filter:
        if dolphie.use_performance_schema:
            where_clause.append("processlist_user = '%s'" % dolphie.user_filter)
        else:
            where_clause.append("User = '%s'" % dolphie.user_filter)

    # Filter database
    if dolphie.db_filter:
        if dolphie.use_performance_schema:
            where_clause.append("processlist_db = '%s'" % dolphie.db_filter)
        else:
            where_clause.append("db = '%s'" % dolphie.db_filter)

    # Filter hostname/IP
    if dolphie.host_filter:
        # Have to use LIKE since there's a port at the end
        if dolphie.use_performance_schema:
            where_clause.append("processlist_host LIKE '%s%%'" % dolphie.host_filter)
        else:
            where_clause.append("Host LIKE '%s%%'" % dolphie.host_filter)

    # Filter time
    if dolphie.query_time_filter:
        if dolphie.use_performance_schema:
            where_clause.append("processlist_time >= '%s'" % dolphie.query_time_filter)
        else:
            where_clause.append("Time >= '%s'" % dolphie.query_time_filter)

    # Filter query
    if dolphie.query_filter:
        if dolphie.use_performance_schema:
            where_clause.append(
                "(processlist_info LIKE '%%%s%%' OR trx_query LIKE '%%%s%%')"
                % (dolphie.query_filter, dolphie.query_filter),
            )
        else:
            where_clause.append("Info LIKE '%%%s%%'" % dolphie.query_filter)

    # Add in our dynamic WHERE clause for filtering
    if where_clause:
        processlist_query = processlist_query.replace("$1", "AND " + " AND ".join(where_clause))
    else:
        processlist_query = processlist_query.replace("$1", "")

    processlist_threads = {}
    # Run the processlist query
    dolphie.main_db_connection.execute(processlist_query)
    threads = dolphie.main_db_connection.fetchall()

    for thread in threads:
        # Don't include Dolphie's threads
        if (
            dolphie.main_db_connection.connection_id == thread["id"]
            or dolphie.secondary_db_connection.connection_id == thread["id"]
        ):
            continue

        # Use trx_query over Performance Schema query since it's more accurate
        if dolphie.use_performance_schema and thread["trx_query"]:
            thread["query"] = thread["trx_query"]
        else:
            thread["query"] = thread["query"]

        host = thread["host"].split(":")[0]
        thread["host"] = dolphie.get_hostname(host)

        processlist_threads[str(thread["id"])] = ProcesslistThread(thread)

    return processlist_threads
